
import numpy as np
import pyvista as pv
import os
import sys
currentDirectory = os.getcwd()
sys.path.insert(1, (currentDirectory + '/library'))
from .MathShapesClass import *
from .PhysicsClass import *

class surf():
    """surf.
        SURF CLASS ATRIBUTES:

          surf.Name = ""
             Name of the element.


          surf.NamePos = (0,0)
             “Name” position in the 2D diagram.


          surf.Note = "None"
             Useful for adding user notes to a surface.


          surf.Rc = 0
             Paraxial radius of curvature in millimeters.


          surf.Cylinder_Rxy_Ratio = 1
             Ratio between the axial and sagittal
             radius of curvature.


          surf.Axicon = 0
             values other than zero an axicon is
             generated with the angle defined


          surf.Thickness = 0.0
             Distance between this surface and the next surface.


          surf.Diameter = 1.0
             Outside diameter of the surface.


          surf.InDiameter = 0.0
             Internal diameter of the surface.


          surf.k = 0.0
             Conicity constant for classical conic surfaces
             k = 0 for spherical
             k = -1 for parabola, etc.
             Default value: 0.0


          surf.DespX = 0.0
          surf.DespY = 0.0
          surf.DespZ = 0.0
            Displacement of the surface in the X, Y and Z axis


          surf.TiltX = 0.0
          surf.TiltY = 0.0
          surf.TiltZ = 0.0
             Rotation of the surface in the X, Y and Z axis


          surf.Order = 0
             Define the order of the transformations.


          surf.AxisMove = 1
             Defines what will happen to the optical axis
             after a coordinate transformation.


          surf.Diff_Ord = 0.0
             Diffraction order.


          surf.Grating_D = 0.0
             Separation between the lines of the diffraction grating.


          surf.Grating_Angle = 0.0
             Angle of the grating lines in the plane of the surface


          surf.ZNK = np.zeros ()
             Zernike polynomials coefficients


          surf.ShiftX = 0
          surf.ShiftY = 0
             Offset the surface function on the X or Y axis.


          surf.Mask = 0
             (0) Do not apply mask,
             (1) Use mask as aperture,
             (2) Use mask as obstruction.
             Default value: 0


          surf.Mask_Shape = Object_3D
             Form of the mask to apply on surface


          surf.AspherData = np.zeros ()
             Asphericity coefficients.


          surf.ExtraData = [f, coef]
             User-defined function for optical surface


          surf.Error_map = [X, Y, Z, SPACE]
             Error map array


          surf.Drawing = 1
             1 for drawn in the 3D plot, 0 to omit.


          surf.Color = [0,0,0]
             Element color for 3D Plot. [R,G,B]


          surf.Solid_3d_stl = "None"
             Path to the 3D solid STL file.

        """


    def __init__(self, Rc = 0.0, Thickness = 1e-11,Diameter = 1.0, InDiameter = 0.0, k = 0.0, Glass = 'AIR'\
        ,ZNK = np.zeros(36)\
        ,DespX = 0.0\
        ,DespY = 0.0\
        ,DespZ = 0.0\
        ,TiltX = 0.0\
        ,TiltY = 0.0\
        ,TiltZ = 0.0\
        ,Order = 0.0\
        ,AxisMove = 1.0\
        ,Diff_Ord = 0.0\
        ,Grating_D = 0.0\
        ,Grating_Angle = 0.0\
        ,ShiftX = 0.0\
        ,ShiftY = 0.0\
        ,Mask_Type = 0.0\

        ,Mask_Shape = "None"\
        ,Solid_3d_stl = "None"\
        ,Cylinder_Rxy_Ratio = 1.0\
        ,Axicon = 0.0\
        ,AspherData = np.zeros(200)\
        ,ExtraData = np.zeros(200)\
        ,Thin_Lens = 0.0\
        ,Name = ''\
        ,Nm_Pos = (0.0, 0.0)\
        ,Note = "None"\
        ,Drawing = 1.0\
        ,Color = [0, 0, 0]\
        ,Error_map = []\
        ,Res = 1\
        ,Surface_type = 0.0\
        ,SURF_FUNC = [conic__surf(0.0, 0.0, 1.0)]\
        ,SPECIAL_SURF_FUNC = []\
        ,SubAperture = [1,0,0], Coating = [[],[],[],[]]):

        # coating = [[R],[A],[W],[THETA]]

        """__init__.
        """
        pass
        self.Rc = Rc
        self.Thickness = Thickness
        self.Diameter = Diameter
        self.InDiameter = InDiameter
        self.k = k
        self.Glass = Glass

        self.ZNK = ZNK
        self.DespX = DespX
        self.DespY = DespY
        self.DespZ = DespZ
        self.TiltX = TiltX
        self.TiltY = TiltY
        self.TiltZ = TiltZ
        self.Order = Order
        self.AxisMove = AxisMove
        self.Diff_Ord = Diff_Ord
        self.Grating_D = Grating_D
        self.Grating_Angle = Grating_Angle
        self.ShiftX = ShiftX
        self.ShiftY = ShiftY
        self.Mask_Type = Mask_Type

        self.Mask_Shape = Mask_Shape
        self.Solid_3d_stl = Solid_3d_stl
        self.Cylinder_Rxy_Ratio = Cylinder_Rxy_Ratio
        self.Axicon = Axicon
        self.AspherData = AspherData
        self.ExtraData = ExtraData
        self.Thin_Lens = Thin_Lens
        self.Name = Name
        self.Nm_Pos = Nm_Pos
        self.Note = Note
        self.Drawing = Drawing
        self.Color = Color
        self.Error_map = Error_map
        self.Res = Res
        self.Surface_type = Surface_type
        self.SURF_FUNC = SURF_FUNC
        self.SPECIAL_SURF_FUNC = SPECIAL_SURF_FUNC
        self.General_Status = self.update()
        self.SubAperture = SubAperture
        self.Coating = Coating

    def RestoreVTK(self):
        Objeto_3D = pv.Disc(center=[0.0, 0.0, 0.0], inner=0, outer=0.001, normal=(0, 0, 1), r_res=3, c_res=3)
        Mask = pv.MultiBlock()
        Mask.append(Objeto_3D)
        self.Mask_Shape = Mask

    def EraseVTK(self):

        self.Mask_Shape = "None"



    def warning(self):
        """warning.
        """
        print(' Warning, two types of surface are set at the same time,')
        print('correct this, dont set any combination que between diffraction order,')
        print(' Thin_Lens and  Solid_3d_stl.')
        print('Default value is defined to: Surface_type=0, ')
        print('    ')
        print('If it is not used, no value should be assigned')
        return

    def build_surface_function(self):
        """build_surface_function.
        """
        self.SURF_FUNC = []
        self.SPECIAL_SURF_FUNC = []
        self.Surface_type = 0
        if (self.Diff_Ord != 0):
            self.Surface_type = 1
            # print('Surface type 1')
        if (self.Thin_Lens != 0):
            self.Surface_type = 2
            # print('Surface type 2')
        if (self.Solid_3d_stl != 'None'):
            self.Surface_type = 3
            # print('Surface type 3')
        if ((self.Diff_Ord != 0) and (self.Thin_Lens != 0)):
            self.warning()
            self.Surface_type = 0
        if ((self.Solid_3d_stl != 'None') and (self.Thin_Lens != 0)):
            self.warning()
            self.Surface_type = 0
        if ((self.Solid_3d_stl != 'None') and (self.Diff_Ord != 0)):
            self.warning()
            self.Surface_type = 0
        if (self.Surface_type == 0):
            self.PHYSICS = snell_refraction_vector_physics()
            if np.any((self.ZNK != 0)):
                NC = len(self.ZNK)
                (self.Zern_pol, self.z_pow) = zernike_expand(NC)
                FUNC_0 = zernike__surf(self.ZNK, self.Zern_pol, self.z_pow, self.Diameter)
                self.SURF_FUNC.append(FUNC_0)
            if np.any((self.AspherData != 0)):
                FUNC_1 = aspheric__surf(self.AspherData)
                self.SURF_FUNC.append(FUNC_1)
            if (self.Rc != 0):
                FUNC_2 = conic__surf(self.Rc, self.k, self.Cylinder_Rxy_Ratio)
                self.SURF_FUNC.append(FUNC_2)
            if (self.Axicon != 0):
                FUNC_3 = axicon__surf(self.Cylinder_Rxy_Ratio, self.Axicon)
                self.SURF_FUNC.append(FUNC_3)
            if np.any((self.ExtraData != 0)):
                FUNC_4 = extra__surf(self.ExtraData)
                self.SURF_FUNC.append(FUNC_4)
            if (len(self.Error_map) != 0):
                [X, Y, Z, SPACE] = self.Error_map
                FUNC_5 = error_map__surf(X, Y, Z, SPACE)
                self.SURF_FUNC.append(FUNC_5)
        if (self.Surface_type == 1):
            self.PHYSICS = diffraction_grating_physics()
            FUNC = conic__surf(0, self.k, self.Cylinder_Rxy_Ratio)
            self.SURF_FUNC.append(FUNC)
        if (self.Surface_type == 2):
            self.PHYSICS = paraxial_exact_physics()
            FUNC = conic__surf(0, self.k, self.Cylinder_Rxy_Ratio)
            self.SURF_FUNC.append(FUNC)
            self.Glass = 'AIR'
        if (self.Surface_type == 3):
            self.PHYSICS = snell_refraction_vector_physics()
            FUNC = conic__surf(0, self.k, self.Cylinder_Rxy_Ratio)
            self.SURF_FUNC.append(FUNC)

    def sigma_z(self, x, y, case):
        """sigma_z.

        Parameters
        ----------
        x :
            x
        y :
            y
        case :
            case
        """
        x = (x + self.ShiftX)
        y = (y + self.ShiftY)
        # Z = 0.0 * np.copy(x)
        con = -1
        N_FUNC = len(self.SURF_FUNC)

        for i in range(0, (N_FUNC - case)):

            if con==-1:
                Z = self.SURF_FUNC[i].calculate(x, y)
                con = 0

            else:
                Z = (self.SURF_FUNC[i].calculate(x, y) + Z)
                con=1

        if con == -1:
            Z = 0.0 * np.copy(x)

        return Z



    def update(self):
        """update.
        """
        self.General_Status = np.asarray([id(self.Rc), id(self.Diameter), id(self.InDiameter), id(self.k), id(self.ZNK), id(self.AspherData), id(self.ExtraData), id(self.Cylinder_Rxy_Ratio), id(self.ShiftX), id(self.ShiftY), id(self.Axicon), id(self.Thin_Lens), id(self.Error_map), id(self.Thickness), id(self.DespX), id(self.DespY), id(self.DespZ), id(self.TiltX), id(self.TiltY), id(self.TiltZ), id(self.Order), id(self.AxisMove), id(self.Glass), id(self.Diff_Ord), id(self.Grating_D), id(self.Grating_Angle), id(self.Mask_Type), id(self.Mask_Shape), id(self.Solid_3d_stl), id(self.Name), id(self.Nm_Pos), id(self.Note), id(self.Drawing), id(self.Color), id(self.Surface_type), id(self.Res)])
        return self.General_Status

    def SaveSetup(self):
        """SaveSetup.
        """
        self.Sv_Rc = self.Rc
        self.Sv_Thickness = self.Thickness
        self.Sv_Diameter = self.Diameter
        self.Sv_InDiameter = self.InDiameter
        self.Sv_k = self.k
        self.Sv_ZNK = self.ZNK
        self.Sv_Glass = self.Glass
        self.Sv_DespX = self.DespX
        self.Sv_DespY = self.DespY
        self.Sv_DespZ = self.DespZ
        self.Sv_TiltX = self.TiltX
        self.Sv_TiltY = self.TiltY
        self.Sv_TiltZ = self.TiltZ
        self.Sv_Order = self.Order
        self.Sv_AxisMove = self.AxisMove
        self.Sv_Diff_Ord = self.Diff_Ord
        self.Sv_Grating_D = self.Grating_D
        self.Sv_Grating_Angle = self.Grating_Angle
        self.Sv_ShiftX = self.ShiftX
        self.Sv_ShiftY = self.ShiftY
        self.Sv_Mask_Type = self.Mask_Type
        self.Sv_Cylinder_Rxy_Ratio = self.Cylinder_Rxy_Ratio
        self.Sv_Axicon = self.Axicon
        self.Sv_AspherData = self.AspherData
        self.Sv_ExtraData = self.ExtraData
        self.Sv_Thin_Lens = self.Thin_Lens
        self.Sv_Error_map = self.Error_map
        self.Sv_SubAperture = self.SubAperture
        self.Sv_Coating = self.Coating

    def RestoreSetup(self):
        """RestoreSetup.
        """
        self.Rc = self.Sv_Rc
        self.Thickness = self.Sv_Thickness
        self.Diameter = self.Sv_Diameter
        self.InDiameter = self.Sv_InDiameter
        self.k = self.Sv_k
        self.ZNK = self.Sv_ZNK
        self.Glass = self.Sv_Glass
        self.DespX = self.Sv_DespX
        self.DespY = self.Sv_DespY
        self.DespZ = self.Sv_DespZ
        self.TiltX = self.Sv_TiltX
        self.TiltY = self.Sv_TiltY
        self.TiltZ = self.Sv_TiltZ
        self.Order = self.Sv_Order
        self.AxisMove = self.Sv_AxisMove
        self.Diff_Ord = self.Sv_Diff_Ord
        self.Grating_D = self.Sv_Grating_D
        self.Grating_Angle = self.Sv_Grating_Angle
        self.ShiftX = self.Sv_ShiftX
        self.ShiftY = self.Sv_ShiftY
        self.Mask_Type = self.Sv_Mask_Type
        self.Cylinder_Rxy_Ratio = self.Sv_Cylinder_Rxy_Ratio
        self.Axicon = self.Sv_Axicon
        self.AspherData = self.Sv_AspherData
        self.ExtraData = self.Sv_ExtraData
        self.Thin_Lens = self.Sv_Thin_Lens
        self.Error_map = self.Sv_Error_map
        self.SubAperture = self.Sv_SubAperture
        self.Coating = self.Sv_Coating



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Perfect Lens"""

import time
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos
from scipy.optimize import fsolve
import scipy

def MyPlot(RK,surf, figure= "Spot", mk=["x"], col = [[0.8,0.0,0.0]]):
    fig = plt.figure(figure)
    ax = fig.add_subplot()
    i=0
    for rk in RK:
        x,y,z,l,m,n = rk.pick(surf[i], coordinates="local")
        plt.scatter(x, y,marker = mk[i], color = col[i])
        i = i + 1
    # https://matplotlib.org/3.1.1/api/markers_api.html
    ax.set_aspect('equal', adjustable='box')
    ax.set_xlabel('x (mm)')
    ax.set_ylabel('y (mm)')
    ax.set_title('Spot diagram')
    plt.show()


def R_RMS_delta(Z1, L, M, N, X0, Y0):
    X1 = ((L / N) * Z1) + X0
    Y1 = ((M / N) * Z1) + Y0
    cenX = np.mean(X1)
    cenY = np.mean(Y1)
    x1 = (X1 - cenX)
    y1 = (Y1 - cenY)
    R2 = ((x1 * x1) + (y1 * y1))
    R_RMS = np.sqrt(np.mean(R2))
    return R_RMS

def BestFocus(X, Y, Z, L, M, N, system, mod=1):
    delta_Z = 0
    ZZ = (L, M, N, X, Y)
    v = scipy.optimize.fsolve(R_RMS_delta, delta_Z, args=ZZ)
    if mod ==1:
        system.SDT[-2].Thickness = system.SDT[-2].Thickness + v[0]
        system.SetData()
        system.SetSolid()
    return system, v[0]


def BestRMS(system,raykeeper):
    x,y,z,l,m,n = raykeeper.pick(-1 , coordinates="local")
    system, deltaZ = BestFocus(x, y, z, l, m, n, system, mod = 0)
    rms = R_RMS_delta(deltaZ,l,m,n,x,y)
    return rms


class Function2Optimize:
    def __init__(self, system, raykeeper, effl, Surf, w, AperVal, AperType):

        self.system = system
        self.w = w
        self.effl = effl
        self.raykeeper = raykeeper

        self.P = Kos.PupilCalc(Lens, Surf, w, AperType, AperVal)

        self.P.Samp, self.P.Ptype, self.P.FieldType = 5, "hexapolar", "angle"
        self.P.FieldX, self.P.FieldY = 0.0, 0.0
        self.ProcessPattern2Field()

    def ProcessPattern2Field(self):
        self.x, self.y, self.z, self.L, self.M, self.N = self.P.Pattern2Field()

    def EFFL(self, V):
        self.system.SDT[1].Rc = V[0]
        self.system.SDT[2].Rc = V[1]


        self.system.SetData()
        self.system.Parax(self.w)
        Kos.TraceLoop(self.x, self.y, self.z, self.L, self.M, self.N, self.w, self.raykeeper, clean = 1)

        D_EFFL = self.system.EFFL - self.effl

        B = BestRMS(self.system, self.raykeeper)

        self.system.RestoreData()
        return [D_EFFL, B]




P_Obj = Kos.surf()
P_Obj.Thickness = 100
P_Obj.Diameter = 120.0

Fa = Kos.surf()
Fa.Rc = 0.0
Fa.Thickness = 12.0
Fa.Glass = "BK7"
Fa.Diameter = 120.0

Fb = Kos.surf()
Fb.Rc = 0.0
Fb.Thickness = 1000.0
Fb.Diameter = 120
Fb.Glass = "AIR"


P_Ima = Kos.surf()
P_Ima.Thickness = 0.0
P_Ima.Diameter = 120.0
P_Ima.Name = "Image plane"
P_Ima.Nm_Pos = (-10,10)
P_Ima.Glass = "AIR"

config_1 = Kos.Setup()

A = [P_Obj, Fa, Fb, P_Ima]
Lens = Kos.system(A, config_1)

Rays = Kos.raykeeper(Lens)







Wave = 0.5
effl = 1000
Lens.Parax(0.5)
Surf=1

print("Initial effective focal length: ", Lens.EFFL)
# print("Initial RMS radius in best focus: ", BestRMS(Lens, Wave))

MyFun = Function2Optimize(Lens, Rays, effl, Surf, Wave, Fa.Diameter, "EPD")

R0 =  -1000000.
R1 =  1000000.
R = scipy.optimize.least_squares(MyFun.EFFL, [R0, R1])


# [R0,R1,K1, K2] = R.x

[R0,R1] = R.x

print("Resulting R0: ", R0, " R1: ", R1)
# print("Resulting R0: ", R0, " R1: ", R1, " K1: ", K1, " K2: ", K2)
Lens.SDT[1].Rc = R0
# Lens.SDT[1].k = K1
Lens.SDT[2].Rc = R1
# Lens.SDT[2].k = K2
Lens.SetData()
Lens.SetSolid()


Surf, W, AperVal, AperType = 1, 0.45, Fa.Diameter, "EPD"
P = Kos.PupilCalc(Lens, Surf, W, AperType, AperVal)
P.Samp, P.Ptype, P.FieldX, P.FieldType = 8, "fanx", 0, "angle"
x, y, z, L, M, N = P.Pattern2Field()

Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)


Lens.Parax(0.5)
print("Final effective focal length: ", Lens.EFFL)



x, y, z, l, m, n = Rays.pick(-1 , coordinates="local")
system, deltaZ = BestFocus(x, y, z, l, m, n, Lens, mod = 1)

x, y, z, L, M, N = P.Pattern2Field()
Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)

x, y, z, l, m, n = Rays.pick(-1 , coordinates="local")
rms = R_RMS_delta(deltaZ,l,m,n,x,y)
print("Final RMS radius: ", rms)




Kos.display2d(Lens, Rays, 1)




P.Samp, P.Ptype, P.FieldX, P.FieldType = 7, "hexapolar", 0, "angle"
x, y, z, L, M, N = P.Pattern2Field()
Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)
RAYS = [Rays]
MK = ["x"]
COL = [[0.8,0.0,0.0]]
SURF = [-1]
MyPlot(RAYS, SURF, figure= "Spot2", mk = MK, col = COL)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Perfect Lens"""

import time
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

start_time = time.time()

# ______________________________________#

P_Obj = Kos.surf(Thickness = 100, Diameter = 30.0)
P_Obj.Name = "o"

L = Kos.surf()
L.Thin_Lens = 50.0
L.Thickness = 50.0
L.Diameter = 30.0

S = Kos.surf(Thickness = 50.0, Diameter = 30.0)
S.Name = "F"

P_Ima = Kos.surf(Diameter = 40.0, Name = "i")

A = [P_Obj, L, S, P_Ima]
config_1 = Kos.Setup()

Lens = Kos.system(A, config_1)
Rays = Kos.raykeeper(Lens)


Surf = 1
W = 0.45
AperVal = L.Diameter
AperType = "EPD"
Pupil = Kos.PupilCalc(Lens, Surf, W, AperType, AperVal)
Pupil.Samp = 1
Pupil.Ptype = "fanx"
Pupil.FieldX = 0.0
Pupil.FieldY = 0.0

# """ FieldType = angle or height """
# Pupil.FieldType = "angle"

# x, y, z, L, M, N = Pupil.Pattern2Field()

# for i in range(0, len(x)):
#     pSource_0 = [x[i], y[i], z[i]]
#     dCos = [L[i], M[i], N[i]]
#     Lens.Trace(pSource_0, dCos, W)
#     Rays.push()

# Kos.display2d(Lens, Rays, 1, arrow=1)






""" FieldType = angle or height """
Pupil.FieldType = "height"

x, y, z, L, M, N = Pupil.Pattern2Field()

for i in range(0, len(x)):
    pSource_0 = [x[i], y[i], z[i]]
    dCos = [L[i], M[i], N[i]]
    Lens.Trace(pSource_0, dCos, W)
    Rays.push()

Kos.display2d(Lens, Rays, 1, arrow=1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Perfect Lens"""

import time
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

import scipy

def R_RMS_delta(Z1, L, M, N, X0, Y0):
    X1 = ((L / N) * Z1) + X0
    Y1 = ((M / N) * Z1) + Y0
    cenX = np.mean(X1)
    cenY = np.mean(Y1)
    x1 = (X1 - cenX)
    y1 = (Y1 - cenY)
    R2 = ((x1 * x1) + (y1 * y1))
    R_RMS = np.sqrt(np.mean(R2))
    return R_RMS

def BestFocus(X, Y, Z, L, M, N, system, mod=1):
    delta_Z = 0
    ZZ = (L, M, N, X, Y)
    v = scipy.optimize.fsolve(R_RMS_delta, delta_Z, args=ZZ)
    if mod ==1:
        system.SDT[-2].Thickness = system.SDT[-2].Thickness + v[0]
        system.SetData()
        system.SetSolid()
    return system, v[0]



P_Obj = Kos.surf(Thickness = 120, Diameter = 120)
L1 = Kos.surf(Thin_Lens = 1000.0, Thickness =1000.0 + 20, Diameter = 120)
L1.Name = "Objetive"
L1.Nm_Pos = (-20,100)
L2 = Kos.surf(Thin_Lens = 20.0, Thickness = 50.0, Diameter = 20.0)
L2.Name = "Eyepiece"
# L2.Nm_Pos = (-20,100)
P_Ima = Kos.surf(Diameter = 60.0, Name = "i")

A = [P_Obj, L1, L2, P_Ima]
config_1 = Kos.Setup()

Lens = Kos.system(A, config_1)
Rays = Kos.raykeeper(Lens)

Surf, W, AperVal, AperType = 1, 0.45, L1.Diameter, "EPD"
P = Kos.PupilCalc(Lens, Surf, W, AperType, AperVal)
P.Samp, P.Ptype, P.FieldType = 1, "fanx", "angle"

P.FieldX = 0.25
x, y, z, L, M, N = P.Pattern2Field()
Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 0)

P.FieldX = 0.0
x, y, z, L, M, N = P.Pattern2Field()
Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 0)


P.FieldX = -0.25
x, y, z, L, M, N = P.Pattern2Field()
Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 0)

Kos.display2d(Lens, Rays, 1, arrow=1)


x,y,z,l,m,n = Rays.pick(-1, coordinates="global")
print(l)
print(m)
print(n)










#from numba import jit


import numpy as np


# @jit(forceobj=True)
def FresnelEnergy(vidrio, NP, NC, ImpVec, SurfNorm, ResVec, SETUP, Wave):
    """FresnelEnergy.

    Parameters
    ----------
    vidrio :
        vidrio
    NP :
        NP
    NC :
        NC
    ImpVec :
        ImpVec
    SurfNorm :
        SurfNorm
    ResVec :
        ResVec
    SETUP :
        SETUP
    Wave :
        Wave
    """
    # global Ts, Tp, Rs, Rp
    if (vidrio != 'MIRROR'):
        (Rp, Rs, Tp, Ts) = fresnel_dielectric(NP, NC, ImpVec, SurfNorm, ResVec)
    if (vidrio == 'MIRROR'):
        n_metal = np.interp(Wave, SETUP.W_alum, SETUP.N_alum)
        k_complex = np.interp(Wave, SETUP.W_alum, SETUP.K_alum)
        (Rp, Rs, Tp, Ts) = fresnel_metal(NP, n_metal, k_complex, ImpVec, SurfNorm)



    # print(self.SDT[j].Coating, ang)
    return (Rp, Rs, Tp, Ts)


# @jit(forceobj=True)
def fresnel_dielectric(NP, NC, LMN_Inc, LMN_nor_surf, LMN_result):
    """fresnel_dielectric.

    Parameters
    ----------
    NP :
        NP
    NC :
        NC
    LMN_Inc :
        LMN_Inc
    LMN_nor_surf :
        LMN_nor_surf
    LMN_result :
        LMN_result
    """
    AA=(LMN_Inc[0]*LMN_nor_surf[0])+(LMN_Inc[1]*LMN_nor_surf[1])+(LMN_Inc[2]*LMN_nor_surf[2])
    CosTheta0 = np.abs(AA)
    BB=(LMN_result[0]*LMN_nor_surf[0])+(LMN_result[1]*LMN_nor_surf[1])+(LMN_result[2]*LMN_nor_surf[2])
    CosTheta1 = np.abs(BB)
    n0 = NP
    n1 = NC
    rs = (((n0 * CosTheta0) - (n1 * CosTheta1)) / ((n0 * CosTheta0) + (n1 * CosTheta1)))
    rp = (((n1 * CosTheta0) - (n0 * CosTheta1)) / ((n1 * CosTheta0) + (n0 * CosTheta1)))

    Rs = (rs ** 2.)
    Rp = (rp ** 2.)
    Tp = (1.0 - Rp)
    Ts = (1.0 - Rs)

    return (Rp, Rs, Tp, Ts)




# @jit(forceobj=True)
def fresnel_metal(NP, n_metal, k_complex, LMN_Inc, LMN_nor_surf):
    """fresnel_metal.

    Parameters
    ----------
    NP :
        NP
    n_metal :
        n_metal
    k_complex :
        k_complex
    LMN_Inc :
        LMN_Inc
    LMN_nor_surf :
        LMN_nor_surf
    """
    n1 = NP
    n2 = np.complex(n_metal, k_complex)
    CosTheta0 = np.abs(np.dot(LMN_Inc, LMN_nor_surf))
    if CosTheta0> 1.0:

        CosTheta0 =1.0
    Thetai = np.arccos(CosTheta0)
    Thetat = np.arcsin(((n1 / n2) * np.sin(Thetai)))
    rs = (((n1 * np.cos(Thetai)) - (n2 * np.cos(Thetat))) / ((n1 * np.cos(Thetai)) + (n2 * np.cos(Thetat))))
    rp = (((n2 * np.cos(Thetai)) - (n1 * np.cos(Thetat))) / ((n1 * np.cos(Thetat)) + (n2 * np.cos(Thetai))))
    Rs = (np.abs(rs) ** 2)
    Rp = (np.abs(rp) ** 2)
    Tp = (1 - Rp)
    Ts = (1 - Rs)
    return (Rp, Rs, Tp, Ts)


# @jit(forceobj=True)
def n_wave_dispersion(krakenSetup, GLSS, Wave):
    """n_wave_dispersion.

    Parameters
    ----------
    krakenSetup :
        krakenSetup
    GLSS :
        GLSS
    Wave :
        Wave
    """
    if (GLSS[0:4] != 'AIR_'):
        CAT = krakenSetup.CAT
        NAMES = krakenSetup.NAMES
        NM = krakenSetup.NM
        ED = krakenSetup.ED
        CD = krakenSetup.CD
        TD = krakenSetup.TD
        OD = krakenSetup.OD
        LD = krakenSetup.LD
        IT = krakenSetup.IT
        if (GLSS == 'MIRROR'):
            n = (- 1.0)
            Alpha = 0.0

        if (GLSS == 'AIR'):
            n = 1.0
            Alpha = 0.0
        if ((GLSS != 'MIRROR') and (GLSS != 'AIR')):
            r = np.argwhere((NAMES == GLSS))[0][0]
            [Dispersion_Formula, MIL, Nd, Vd, Exclude_sub, Status] = NM[r]
            C = CD[r]
            lam2 = Wave **2.

            if (Dispersion_Formula == 1):
                """ Schott """
                (a0, a1, a2, a3, a4, a5) = (C[0], C[1], C[2], C[3], C[4], C[5])
                lam4 = Wave **4.
                lam6 = Wave **6.
                lam8 = Wave **8.
                n2 = (((((a0 + (a1 * lam2)) + (a2 / lam2)) + (a3 / lam4)) + (a4 / lam6)) + (a5 / lam8))
                n = np.sqrt(n2)

            if (Dispersion_Formula == 2):
                """ Sellmeier1 """
                (K1, L1, K2, L2, K3, L3) = (C[0], C[1], C[2], C[3], C[4], C[5])
                p1 = ((K1 * lam2) / (lam2 - L1))
                p2 = ((K2 * lam2) / (lam2 - L2))
                p3 = ((K3 * lam2) / (lam2 - L3))
                n = np.sqrt((((p1 + p2) + p3) + 1.0))

            if (Dispersion_Formula == 3):
                """Herzberger"""
                CA = C[0]
                CB = C[1]
                CC = C[2]
                CD = C[3]
                CE = C[4]
                CF = C[5]
                CL = (1.0 / ((Wave * 2.0) - 0.028))
                n = (((((CA + (CB * CL)) + (CC * (CL ** 2.0))) + (CD * (Wave ** 2.0))) + (CE * (Wave ** 4.0))) + (CF * (Wave ** 6.0)))
            if (Dispersion_Formula == 4):
                """Sellmeier 2"""
                GLSN = C[0] + (C[1] * Wave**2 / (Wave**2 - (C[2])**2)) + (C[3] * Wave**2 / (Wave**2 - (C[4])**2))
                n = np.sqrt(GLSN + 1.0)

            if (Dispersion_Formula == 5):
                """Conrady"""
                n = C[0] + (C[1] / Wave) + (C[2] / Wave**3.5)

            if (Dispersion_Formula == 6):
                """Sellmeier 3"""
                GLSN = (C[0] * Wave**2 / (Wave**2 - C[1])) + (C[2] * Wave**2 / (Wave**2 - C[3])) + (C[4] * Wave**2 / (Wave**2 - C[5])) + (C[6] * Wave**2 / (Wave**2 - C[7]))
                n = np.sqrt(GLSN + 1.0)

            if (Dispersion_Formula == 7):
                """Hanbook of Optics 1"""
                GLSN = C[0] + (C[1] / (Wave**2 - C[2])) - (C[3] * Wave**2)
                n = np.sqrt(GLSN)

            if (Dispersion_Formula == 8):
                """Hanbook of Optics 2"""
                GLSN = C[0] + (C[1] * Wave**2 / (Wave**2 - C[2])) - (C[3] * Wave**2)
                n = np.sqrt(GLSN)

            if (Dispersion_Formula == 9):
                """Sellmeier 4"""
                GLSN = C[0] + (C[1] * Wave**2 / (Wave**2 - C[2])) + (C[3] * Wave**2 / (Wave**2 - C[4]))
                n = np.sqrt(GLSN)

            if (Dispersion_Formula == 10):
                """Extended"""
                GLSN = C[0] + (C[1] * Wave**2) + (C[2] * Wave**-2) + (C[3] * Wave**-4) + (C[4] * Wave**-6) + \
                              (C[5] * Wave**-8) + (C[6] * Wave**-10) + (C[7] * Wave**-12)
                n = np.sqrt(GLSN)

            if (Dispersion_Formula == 11):
                """Sellmeier 5"""
                GLSN = (C[0] * Wave**2 / (Wave**2 - C[1])) + (C[2] * Wave**2 / (Wave**2 - C[3])) + (C[4] * Wave**2 / (Wave**2 - C[5])) + (C[6] * Wave**2 / (Wave**2 - C[7])) + (C[8] * Wave**2 / (Wave**2 - C[9]))
                n = np.sqrt(GLSN + 1.0)


            if (Dispersion_Formula == 12):
                """Extended 2"""
                GLSN = C[0] + (C[1] * Wave**2) + (C[2] * Wave**-2) + (C[3] * Wave**-4) + (C[4] * Wave**-6) + (C[5] * Wave**-8) + (C[6] * Wave**4) + (C[7] * Wave**6)
                n = np.sqrt(GLSN)


            if (Dispersion_Formula == 13):
                """HIKARI Catalog"""
                GLSN0 = C[0]
                GLSN1 =(C[1] * Wave**2)
                GLSN2 =(C[2] * Wave**4)
                GLSN3 =(C[3] * Wave**-2)
                GLSN4 =(C[4] * Wave**-4)
                GLSN5 =(C[5] * Wave**-6)
                GLSN6 =(C[6] * Wave**-8)
                GLSN7 =(C[7] * Wave**-10)
                GLSN8 =(C[8] * Wave**-12)
                GLSN = GLSN0 + GLSN1 + GLSN2 + GLSN3 + GLSN4 + GLSN5 + GLSN6 + GLSN7 + GLSN8

                """ n(W)2 = A0 + A1*W**2 + A2*W**4 + A3*W**−2 + A4*W**−4 + A5*W**−6 + A6*W**−8 + A7*W**−10 + A8*W**−12"""

                n = np.sqrt(GLSN)



            [Wa, Tr, Th] = IT[r]
            TR = np.interp(Wave, np.asarray(Wa), np.asarray(Tr))
            Alpha = ((- np.log(TR)) / Th[0])




    else:
        n = float(GLSS[4:])
        Alpha = 0.0

    return (n, Alpha)



# @jit(forceobj=True)
def ParaxCalc(N_Prec, SDT, SuTo, n, Gl):
    """ParaxCalc.

    Parameters
    ----------
    N_Prec :
        N_Prec
    SDT :
        SDT
    SuTo :
        SuTo
    n :
        n
    Gl :
        Gl
    """
    j = 0
    PrevN = N_Prec[j]
    S_Matrix = []
    N_Matrix = []
    CC = []
    DD = []
    KK = []
    for j in range(0, n+0):
        sag_rad = (SDT[j].Diameter / 400.0)
        SG = SuTo.SurfaceShape(sag_rad, 0.0, j)
        if (SG != 0.0):
            R_p = ((sag_rad * sag_rad) / (2.0 * SG))
        else:
            R_p = 9999999999999.9
        if (SDT[j].Solid_3d_stl != 'None'):
            R_p = 9999999999999.9
        Glass = Gl[j]
        if (Glass != 'NULL'):
            CurrN = N_Prec[j]
        if (Glass == 'NULL'):
            CurrN = PrevN
        n1 = np.abs(PrevN)
        n2 = np.abs(CurrN)
        dd = (SDT[j].Thickness + SDT[j].DespZ)
        R = R_p
        if (Glass == 'MIRROR'):
            n1 = (- n1)

        if (SDT[j].Thin_Lens == 0):
            RR = np.matrix([[(n1 / n2), ((n1 - n2) / (n2 * R))], [0.0, 1]])
        else:
            RR = np.matrix([[1.0, -1.0 / SDT[j].Thin_Lens], [0.0, 1]])

        CC.append((1 / R))
        DD.append(dd)
        KK.append(0)
        TT = np.matrix([[1.0, 0.0], [dd, 1.0]])
        S_Matrix.append(RR)
        N_Matrix.append(('R surf ' + str(j)))
        S_Matrix.append(TT)
        N_Matrix.append(((('T surf ' + str(j)) + ' to ') + str((j + 1))))
        PrevN = CurrN
    SistemMatrix = np.matrix([[1.0, 0.0], [0.0, 1.0]])
    for Mi in S_Matrix[::(- 1)]:
        SistemMatrix = np.dot(SistemMatrix, Mi)

    """ Hecht E. - Optics-AW (2002) """
    a = SistemMatrix[0][(0, 0)]
    b = SistemMatrix[0][(0, 1)]
    c = SistemMatrix[1][(0, 0)]
    d = SistemMatrix[1][(0, 1)]
    EFFL = ((- 1) / b)
    PPA = ((1 - a) / -b)
    PPP = ((d - 1) / -b)
    CC = np.asarray(CC)
    DD = np.asarray(DD)
    return (SistemMatrix, S_Matrix, N_Matrix, a, b, c, d, EFFL, PPA, PPP, CC, N_Prec, DD)














# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Examp-2M-STL_ImageSlicer.py
"""

import pkg_resources
""" Looking for if KrakenOS is installed, if not, it assumes that
an folder downloaded from github is run"""

required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("Not installed")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos


import numpy as np
import matplotlib.pyplot as plt
import scipy
import os

A1 = 1

if A1 == 0:
    # _________________________________________________________________#
    P_Obj = Kos.surf()
    P_Obj.Rc = 0
    P_Obj.Thickness = 1000 + 3.452200000000000E+003
    P_Obj.Glass = "AIR"
    P_Obj.Diameter = 1.059E+003 * 2.0
    # _________________________________________________________________#
    Thickness = 3.452200000000000E+003
    M1 = Kos.surf()
    M1.Rc = -9.638000000004009E+003
    M1.Thickness = -Thickness
    M1.k = -1.077310000000000E+000
    M1.Glass = "MIRROR"
    M1.Diameter = 1.059E+003 * 2.0
    M1.InDiameter = 250 * 2.0
    # _________________________________________________________________#
    M2 = Kos.surf()
    M2.Rc = -3.93E+003
    M2.Thickness = Thickness + 1.037525880125084E+003
    M2.k = -4.328100000000000E+000
    M2.Glass = "MIRROR"
    M2.Diameter = 3.365E+002 * 2.0
    M2.AxisMove = 0
    # _________________________________________________________________#
    P_Image_A = Kos.surf()
    P_Image_A.Diameter = 10.0
    P_Image_A.Glass = "AIR"
    P_Image_A.Thickness = 10
    P_Image_A.Name = "Image plane Tel"
    P_Image_A.DespZ = -100.0
    # _________________________________________________________________#
    A = [P_Obj, M1, M2, P_Image_A]

    # _________________________________________________________________#
    configuracion_1 = Kos.Setup()
    Telescopio = Kos.system(A, configuracion_1)
    Rayos = Kos.raykeeper(Telescopio)


    # _________________________________________________________________#

    # Gaussian
    def f(x):
        x = np.rad2deg(x)
        seing = 1.2 / 3600.0
        sigma = seing / 2.3548
        mean = 0
        standard_deviation = sigma
        y = scipy.stats.norm(mean, standard_deviation)
        res = y.pdf(x)
        return res


    Sun = Kos.SourceRnd()
    Sun.field = 4 * 1.2 / (2.0 * 3600.0)
    Sun.fun = f
    Sun.dim = 2100
    Sun.num = 100000
    L, M, N, X, Y, Z = Sun.rays()

    Xr = np.zeros_like(L)
    Yr = np.zeros_like(L)
    Zr = np.zeros_like(L)

    Lr = np.zeros_like(L)
    Mr = np.zeros_like(L)
    Nr = np.zeros_like(L)

    NM = np.zeros_like(L)

    con = 0
    con2 = 0
    W = 0.6

    for i in range(0, Sun.num):
        if con2 == 10:
            print(100. * i / Sun.num)
            con2 = 0

        pSource_0 = [X[i], Y[i], Z[i]]
        dCos = [L[i], M[i], N[i]]
        Telescopio.Trace(pSource_0, dCos, W)

        x, y, z = Telescopio.XYZ[-1]
        l, m, n = Telescopio.LMN[-1]
        Xr[con] = x
        Yr[con] = y
        Zr[con] = z
        Lr[con] = l
        Mr[con] = m
        Nr[con] = n

        if Telescopio.NAME[-1] == "Image plane Tel":
            NM[con] = i
        else:
            NM[con] = -1

        con = con + 1
        con2 = con2 + 1
        # Rayos.push()

    args = np.argwhere(NM != -1)

    X = Xr[args]
    Y = Yr[args]
    Z = Zr[args]

    L = Lr[args]
    M = Mr[args]
    N = Nr[args]
    W = W * np.ones_like(N)

    Rays = np.hstack((X, Y, Z, L, M, N, W))
    outfile = "savedRays.npy"
    np.save(outfile, Rays)









################################################################

else:
    P_Obj = Kos.surf()
    P_Obj.Rc = 0
    P_Obj.Thickness = 100. + 0.5
    P_Obj.Glass = "AIR"
    P_Obj.Diameter = 10

    currentDirectory = os.getcwd()
    direc = r"Jherrera-ImageSlicerBW-00.stl"
    P_ImageSlicer = Kos.surf()
    P_ImageSlicer.Diameter = 10.0
    P_ImageSlicer.Glass = "BK7"
    P_ImageSlicer.Name = "Image slicer"
    P_ImageSlicer.Solid_3d_stl = direc
    P_ImageSlicer.Thickness = 13
    P_ImageSlicer.TiltX = 180.0
    P_ImageSlicer.DespX = -0.55
    P_ImageSlicer.DespY = -0.03
    P_ImageSlicer.AxisMove = 0

    P_Ima = Kos.surf()
    P_Ima.Diameter = 10.0
    P_Ima.Glass = "AIR"
    P_Ima.Name = "Plano imagen"

    # _________________________________________________________________#

    A = [P_Obj, P_ImageSlicer, P_Ima]
    configuracion_1 = Kos.Setup()
    ImageSlicer = Kos.system(A, configuracion_1)
    Rayos = Kos.raykeeper(ImageSlicer)

    outfile = "savedRays.npy"
    R = np.load(outfile)
    print(np.shape(R))
    X, Y, Z, L, M, N, W = R[:, 0], R[:, 1], R[:, 2], R[:, 3], R[:, 4], R[:, 5], R[:, 6]

    nrays = 2000
    Xr = np.zeros(nrays)
    Yr = np.zeros(nrays)
    Zr = np.zeros(nrays)
    Lr = np.zeros(nrays)
    Mr = np.zeros(nrays)
    Nr = np.zeros(nrays)
    NM = np.zeros(nrays)

    con = 0
    con2 = 0

    for i in range(0, nrays):
        if con2 == 10:
            print(100. * i / nrays)
            con2 = 0

        pSource_0 = [X[i], Y[i], Z[i] * 0]
        dCos = [L[i], M[i], N[i]]
        ImageSlicer.NsTrace(pSource_0, dCos, W[i])

        x, y, z = ImageSlicer.XYZ[-1]
        l, m, n = ImageSlicer.LMN[-1]
        Xr[con] = x
        Yr[con] = y
        Zr[con] = z
        Lr[con] = l
        Mr[con] = m
        Nr[con] = n

        AA = ImageSlicer.SURFACE
        AA = np.asarray(AA)
        AW = np.argwhere(AA == 1)
        if ImageSlicer.NAME[-1] == "Plano imagen" and len(AW) < 10 and ImageSlicer.TT < 0.9:

            # and ImageSlicer.TT<0.9 and ImageSlicer.TT>0.4

            NM[con] = i
            Rayos.push()
        else:
            NM[con] = -1

        con = con + 1
        con2 = con2 + 1

    args = np.argwhere(NM != -1)

    X = Xr[args]
    Y = Yr[args]
    Z = Zr[args]

    L = Lr[args]
    M = Mr[args]
    N = Nr[args]
    W = W * np.ones_like(N)

    ###################
    plt.plot(X, Y, '.', c="r", markersize=1)

    # axis labeling
    plt.xlabel('x')
    plt.ylabel('y')

    # figure name
    plt.title('Dot Plot')
    plt.axis('square')
    plt.show()

    #             Rays.push()
    Kos.display3d(ImageSlicer, Rayos, 0)



import numpy as np
from .MathShapesClass import *

def RMS_Fitting_Error(SA, X, Y, Z, Zern_pol, z_pow):
    """RMS.

    Parameters
    ----------
    SA :
        SA
    X :
        X
    Y :
        Y
    Z :
        Z
    Zern_pol :
        Zern_pol
    z_pow :
        z_pow
    """
    NZ = []
    SE = np.copy(SA)
    # SE[0] = 0
    # SE[1] = 0
    # SE[2] = 0
    #for i in range(np.shape(X)[0]):
    NZ = Wavefront_Phase(X, Y, SE, Zern_pol, z_pow)
    NZ = np.asarray(NZ)

    # SE[0] = 0
    # SE[1] = 0
    # SE[2] = 0
    # NZZ = Wavefront_Phase(X, Y, SE, Zern_pol, z_pow)
    # print("p2v ", np.abs(np.min(NZZ)-np.max(NZZ)))
    g = (NZ - Z)
    g = (g * g)
    g = np.mean(g)
    error = np.sqrt(g)
    # gg = NZZ-np.mean(NZZ)
    # gg = (gg * gg)
    # gg = np.mean(gg)
    # RMS = np.sqrt(gg)
    return (error)

def Zernike_Fitting(x1, y1, Z1, Arr, minimum=0.000000001):
    """Zernike_Fitting.

    Parameters
    ----------
    x1 :
        x1
    y1 :
        y1
    Z1 :
        Z1
    A :
        A
    minimum :
        minimum
    """

    # Z1=Z1-np.mean(Z1)

    NC = len(Arr)
    (Zern_pol, z_pow) = zernike_expand(NC)
    for i in range(0, 2):
        Zi = System_Matrix_Zernikes(x1, y1, Arr, Zern_pol, z_pow, 0)
        ZT = Zi.T
        ZTZ = np.matmul(ZT, Zi)
        ZTZ_1 = np.linalg.inv(ZTZ)
        ZTZ_1_ZT = np.matmul(ZTZ_1, ZT)
        D = Z1
        D = np.asmatrix(D)
        D = D.T
        SA = np.zeros_like(Arr)
        NA = Arr.shape[0]
        MA = np.asarray(np.matmul(ZTZ_1_ZT, D))



        p = 2.5
        A=Zi
        x=MA
        b=D

        A_T=A.T
        A_T_A=np.matmul(A_T,A)
        Inv_A_T_A=np.linalg.inv(A_T_A)
        Inv_A_T_A_A_T = np.matmul(Inv_A_T_A, A_T)
        x=np.matmul(Inv_A_T_A_A_T, b)


        # for s in range(0,15):
        #     E=np.abs(np.matmul(A,x)-D)
        #     #print(np.sum(np.abs(E)))
        #     E=np.resize(E,(np.shape(E)[0]))
        #     e = np.diag(E)

        #     W=np.power(e, ((p-2.0)/2.0))
        #     WA=np.matmul(W,A)

        #     WAT=WA.T
        #     WATxWA=np.matmul(WAT,WA)
        #     InvWATxWA=np.linalg.inv(WATxWA)
        #     WATW=np.matmul(WAT,W)
        #     f1=np.matmul(InvWATxWA,WATW)
        #     x=np.matmul(f1,D)

        #     print(np.sum(np.abs(E)), " Error")

        MA=x
        # print("-------------")






        cont = 0
        ZZ = []
        for i1 in range(0, Arr.shape[0]):
            ZZ.append(zernike_math_notation(i1, Zern_pol, z_pow))
        ZZ = np.asarray(ZZ)
        for i2 in range(0, NA):
            if (Arr[i2] != 0):
                SA[i2] = MA[cont][0]
                cont = (cont + 1)
            else:
                SA[i2] = 0.0
        Arr = np.abs(SA)
        Zeros = np.argwhere((Arr > minimum))
        AA = np.zeros_like(Arr)
        AA[Zeros] = 1
        Arr = AA
    FITTINGERROR= RMS_Fitting_Error(SA, x1, y1, Z1, Zern_pol, z_pow)
    # print("(RMS) Fitting error: ", FITTINGERROR)

    SE=np.copy(SA)
    SE[0]=0
    RMS2Chief=np.sqrt(np.sum(SE**2.0))
    # print(RMS2Chief, "RMS(to chief) From fitted coefficents")

    SE[1]=0
    SE[2]=0
    RMS2Centroid=np.sqrt(np.sum(SE**2.0))
    # print(RMS2Centroid, "RMS(to centroid) From fitted coefficents")


    return (SA, ZZ, RMS2Chief, RMS2Centroid, FITTINGERROR)

def Wavefront_Zernike_Phase(x, y, COEF):
    """Wavefront_Zernike_Phase.

    Parameters
    ----------
    x :
        x
    y :
        y
    COEF :
        COEF
    """
    NC = len(COEF)
    (Zern_pol, z_pow) = zernike_expand(NC)
    tcoef = COEF.shape[0]
    p = np.sqrt(((x * x) + (y * y)))
    f = np.arctan2(x, y)
    ZFP = 0.0
    for i in range(0, tcoef):
        if (COEF[i] != 0):
            ZFP = (ZFP + (COEF[i] * zernike_polynomials(i, p, f, Zern_pol, z_pow)))
    return ZFP

def Wavefront_Phase(x, y, COEF, Zern_pol, z_pow):
    """Wavefront_Phase.

    Parameters
    ----------
    x :
        x
    y :
        y
    COEF :
        COEF
    Zern_pol :
        Zern_pol
    z_pow :
        z_pow
    """
    tcoef = COEF.shape[0]
    p = np.sqrt(((x * x) + (y * y)))
    f = np.arctan2(x, y)
    ZFP = 0.0
    for i in range(0, tcoef):
        if (COEF[i] != 0):
            ZFP = (ZFP + (COEF[i] * zernike_polynomials(i, p, f, Zern_pol, z_pow)))
    return ZFP

def Wf_XY_Components(x, y, N, Zern_pol, z_pow):
    """Wf_XY_Components.

    Parameters
    ----------
    x :
        x
    y :
        y
    N :
        N
    Zern_pol :
        Zern_pol
    z_pow :
        z_pow
    """
    A = np.zeros(Zern_pol.shape[0])
    A[N] = 1.0
    z = Wavefront_Phase(x, y, A, Zern_pol, z_pow)
    return z

def System_Matrix_Zernikes(x, y, A, Zern_pol, z_pow, fz):
    """System_Matrix_Zernikes.

    Parameters
    ----------
    x :
        x
    y :
        y
    A :
        A
    Zern_pol :
        Zern_pol
    z_pow :
        z_pow
    fz :
        fz
    """
    NA = np.argwhere((A == 1))
    n_NA = NA.shape[0]
    Tp = x.shape[0]
    ZU = np.zeros((Tp, n_NA))
    cont = 0
    for h in range(0, Tp):
        for n in range(0, n_NA):
            F = Wf_XY_Components(x[h], y[h], NA[n], Zern_pol, z_pow)
            ZU[(cont, n)] = F
        cont = (cont + 1)
    return ZU





import numpy as np
from .Physics import *
from numba import jit

# from numba import double, autojit

class snell_refraction_vector_physics():
    """snell_refraction_vector_physics.
    """


    def __init__(self):
        """__init__.
        """
        pass

    def calculate(self, s1_norm, Nsurf_norm, n1, n2, empty1, empty2, empty3, empty4, Secuen):
    #     """calculate.

    #     Parameters
    #     ----------
    #     s1_norm :
    #         s1_norm
    #     Nsurf_norm :
    #         Nsurf_norm
    #     n1 :
    #         n1
    #     n2 :
    #         n2
    #     empty1 :
    #         empty1
    #     empty2 :
    #         empty2
    #     empty3 :
    #         empty3
    #     empty4 :
    #         empty4
    #     Secuen :
    #         Secuen
    #     """

    #     cos = (np.dot(Nsurf_norm, s1_norm))

    #     if (cos < (- 1.0)):
    #         ang = 180
    #     else:
    #         ang = np.rad2deg(np.arccos(cos))

    #     if (ang <= 90.0):
    #         Nsurf_norm = (- Nsurf_norm)

    #     Nsurf_Cros_s1 = np.cross(Nsurf_norm, s1_norm)
    #     SIGN = 1

    #     if (n2 == (- 1)):
    #         n2 = (- n1)
    #         SIGN = (- 1)

    #     NN = (n1 / n2)
    #     d22=np.dot(Nsurf_Cros_s1, Nsurf_Cros_s1)
    #     R = ((NN * NN) * d22)

    #     if (Secuen == 1):
    #         R = 2.0

    #     if (R > 1):
    #         n2 = (- n1)
    #         NN = (n1 / n2)
    #         R = ((NN * NN) * d22)
    #         SIGN = (- 1)

    #     s2 = ((NN * np.cross(Nsurf_norm, np.cross((- Nsurf_norm), s1_norm))) - (Nsurf_norm * np.sqrt((1.0 - ((NN * NN) * d22)))))


    #     cos = np.dot([0, 0, 1], s2)

    #     ang = np.rad2deg(np.arccos(cos))


    #     return (s2, np.abs(n2), SIGN)

        return calculate2(s1_norm, Nsurf_norm, n1, n2, empty1, empty2, empty3, empty4, Secuen)


# @jit(nopython=True)
def calculate2(s1_norm, Nsurf_norm, n1, n2, empty1, empty2, empty3, empty4, Secuen):
    """calculate.

    Parameters
    ----------
    s1_norm :
        s1_norm
    Nsurf_norm :
        Nsurf_norm
    n1 :
        n1
    n2 :
        n2
    empty1 :
        empty1
    empty2 :
        empty2
    empty3 :
        empty3
    empty4 :
        empty4
    Secuen :
        Secuen
r = rayo
r'=reflejado
S=normal superficie
    """


    Nv = np.asarray(Nsurf_norm)
    Iv = np.asarray(s1_norm)


    """ checking if the normal to the surface is in the
    direction of space where the ray is coming from and
    not where it is going """

    cos=np.dot(Nv, Iv)


    if (cos < (- 1.0)):
        ang = 180.0
    else:
        ang = np.rad2deg(np.arccos(cos))

    if (ang <= 90.0):
        Nv = - Nv
    else:
        ang = 180 - ang

    # print(ang)
    """----------------------------------------------"""


    Nsurf_Cros_s1 = np.cross(Nv, Iv)
    SIGN = 1.0

    if (n2 == - 1.0):
        n2 = - n1
        SIGN = -1.0

    NN = (n1 / n2)
    d22=np.dot(Nsurf_Cros_s1, Nsurf_Cros_s1)
    R = (NN * NN) * d22

    if (Secuen == 1.0):
        R = 2.0

    if (R > 1.0):
        # print("Replexion total interna : ", R)
        n2 = - n1
        NN = n1 / n2
        # R = (NN * NN) * d22
        SIGN = -1.0

    c1 = np.dot(Nv,Iv)
    if c1 < 0.0 :
        c1 = np.dot(-Nv,Iv)
    IP = ((NN**2)*(1-(c1**2.0)))

    c2 = np.sqrt(1.0-IP)
    T = NN * Iv + ((( NN * c1 ) - c2)) * Nv

    return T, np.abs(n2), SIGN, ang


# @jit(nopython=True)
# def calculate2(s1_norm, Nsurf_norm, n1, n2, empty1, empty2, empty3, empty4, Secuen):
#     """calculate.

#     Parameters
#     ----------
#     s1_norm :
#         s1_norm
#     Nsurf_norm :
#         Nsurf_norm
#     n1 :
#         n1
#     n2 :
#         n2
#     empty1 :
#         empty1
#     empty2 :
#         empty2
#     empty3 :
#         empty3
#     empty4 :
#         empty4
#     Secuen :
#         Secuen
#     """


#     # Nsurf_norm = 1. * np.asarray(Nsurf_norm1)
#     # s1_norm = 1. * np.asarray(s1_norm1)


#     N1 = [Nsurf_norm[0], Nsurf_norm[1], Nsurf_norm[2]]
#     S1 = [s1_norm[0], s1_norm[1], s1_norm[2]]

#     cos=(N1[0]*S1[0]) + (N1[1]*S1[1]) + (N1[2]*S1[2])

#     # cos = np.dot(N1, S1)

#     if (cos < (- 1.0)):
#         ang = 180.0
#     else:
#         ang = np.rad2deg(np.arccos(cos))

#     if (ang <= 90.0):
#         Nsurf_norm = (- Nsurf_norm)

#     Nsurf_Cros_s1 = np.cross(Nsurf_norm, s1_norm)
#     SIGN = 1.0

#     if (n2 == (- 1.0)):
#         n2 = (- n1)
#         SIGN = (- 1.0)

#     NN = (n1 / n2)
#     d22=np.dot(Nsurf_Cros_s1, Nsurf_Cros_s1)
#     R = ((NN * NN) * d22)

#     if (Secuen == 1.0):
#         R = 2.0

#     if (R > 1.0):
#         n2 = (- n1)
#         NN = (n1 / n2)
#         R = ((NN * NN) * d22)
#         SIGN = (- 1.0)

#     s2 = ((NN * np.cross(Nsurf_norm, np.cross((- Nsurf_norm), s1_norm))) - (Nsurf_norm * np.sqrt((1.0 - ((NN * NN) * d22)))))


#     # cos = np.dot([0.0, 0.0, 1.0], [s2[0], s2[1], s2[2]])
#     # cos = s2[2]



#     # ang = np.rad2deg(np.arccos(cos))


#     return s2, np.abs(n2), SIGN





class paraxial_exact_physics():
    """paraxial_exact_physics.
    """


    def __init__(self):
        """__init__.
        """
        pass

    def calculate(self, s1_norm, Nsurf_norm, n1, n2, empty1, empty2, empty3, empty4, empty5):
        """calculate.

        Parameters
        ----------
        s1_norm :
            s1_norm
        Nsurf_norm :
            Nsurf_norm
        n1 :
            n1
        n2 :
            n2
        empty1 :
            empty1
        empty2 :
            empty2
        empty3 :
            empty3
        empty4 :
            empty4
        empty5 :
            empty5
        """
        return (Nsurf_norm, 1, 1, 0)





class diffraction_grating_physics():
    """diffraction_grating_physics.
    """


    def __init__(self):
        """__init__.
        """
        pass

    def calculate(self, S, R, N, Np, P, Ord, d, W, Secuent):

        cos=np.dot(R,S)
        Ang=np.rad2deg(np.arccos( cos))

        ang = Ang

        if (cos < (- 1.0)):
            ang = 180.0
        else:
            ang = np.rad2deg(np.arccos(cos))

        if (ang >= 90.0):
            ang = 180 - ang


        if np.abs(Ang) < 90:
            R = -R


        """calculate.

        Parameters
        ----------
        S :
            S
        R :
            R
        N :
            N
        Np :
            Np
        D :
            D
        Ord :
            Ord
        d :
            d
        W :
            W
        Secuent :
            Secuent
        """


        D = np.cross(R, P)



        lamb = W
        RefRef = ((- 1.) * np.sign(Np))
        SIGN = 1.

        # print(RefRef, Np)
        if (Np == (- 1.)):
            Np = np.abs(N)

##########################################
        mu = (N / Np)
        T = ((Ord * lamb) / (Np * d))
##########################################

        V = (mu * np.dot(R,S))
        W = ((((mu ** 2.0) - 1.) + (T ** 2.0)) - (((2.0 * mu) * T) * np.dot(D,S)))




        Q1 = ( np.sqrt((V**2 - W))) - V
        Q2 = (-np.sqrt((V**2 - W))) - V

        if RefRef == 1:
            if Q1 > Q2:
                Q = Q1
            else:
                Q = Q2

        if RefRef == -1:
            if Q1 < Q2:
                Q = Q1
            else:
                Q = Q2


        S = np.asarray(S)
        Sp = (((mu * S) - (T * D)) + (Q * R))

        SIGN = SIGN*-1*RefRef
        return (Sp, np.abs(Np), SIGN, ang)






def TraceLoop(x, y, z, L, M, N, W, Container, clean = 1):
    """TraceLoop.

    Parameters
    ----------
    x :
        x
    y :
        y
    z :
        z
    L :
        L
    M :
        M
    N :
        N
    W :
        W
    Container :
        Container
    """
    if clean ==1:
        Container.clean()
    System = Container.SYSTEM
    for i in range(0, len(x)):
        pSource_0 = [x[i], y[i], z[i]]
        dCos = [L[i], M[i], N[i]]
        System.Trace(pSource_0, dCos, W)
        Container.push()
    return 0

def NsTraceLoop(x, y, z, L, M, N, W, Container, clean = 1):
    """NsTraceLoop.

    Parameters
    ----------
    x :
        x
    y :
        y
    z :
        z
    L :
        L
    M :
        M
    N :
        N
    W :
        W
    Container :
        Container
    """
    if clean ==1:
        Container.clean()
    System = Container.SYSTEM
    for i in range(0, len(x)):
        pSource_0 = [x[i], y[i], z[i]]
        dCos = [L[i], M[i], N[i]]
        System.NsTrace(pSource_0, dCos, W)
        Container.push()
    return 0




# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp  TEl 2M Pupila"""

import matplotlib.pyplot as plt
import numpy as np

import sys
sys.path.insert(1, '/Users/joelherreravazquez/Documents/GitHub/Kraken-Optical-Simulator/')

import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0
P_Obj.Thickness = 1000 + 3.452200000000000E+003
P_Obj.Glass = "AIR"
P_Obj.Diameter = 1.059E+003 * 2.0

# ______________________________________#

Thickness = 3.452200000000000E+003
M1 = Kos.surf()
M1.Rc = -9.638000000004009E+003
M1.Thickness = -Thickness
M1.k = -1.077310000000000E+000
M1.Glass = "MIRROR"
M1.Diameter = 1.059E+003 * 2.0
M1.InDiameter = 250 * 2.0

# ______________________________________#

M2 = Kos.surf()
M2.Rc = -3.93E+003
M2.Thickness = Thickness + 1.037525880125084E+003
M2.k = -4.328100000000000E+000
M2.Glass = "MIRROR"
M2.Diameter = 3.365E+002 * 2.0
M2.TiltY = 0.1
M2.TiltX = 0.1
M2.AxisMove = 0

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Diameter = 300.0
P_Ima.Glass = "AIR"
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, M1, M2, P_Ima]
configuracion_1 = Kos.Setup()
Telescopio = Kos.system(A, configuracion_1)

# ______________________________________#

W = 0.4
sup = 1
AperVal = 2010
AperType = "EPD"  # "STOP"
Pup = Kos.PupilCalc(Telescopio, sup, W, AperType, AperVal)

# ______________________________________#

print("Radio pupila de entrada: ")
print(Pup.RadPupInp)
print("Posicion pupila de entrada: ")
print(Pup.PosPupInp)
print("Radio pupila de salida: ")
print(Pup.RadPupOut)
print("Posicion pupila de salida: ")
print(Pup.PosPupOut)
print("Posicion pupila de salida respecto al plano focal: ")
print(Pup.PosPupOutFoc)
print("Orientación pupila de salida")
print(Pup.DirPupSal)
[L, M, N] = Pup.DirPupSal
print(L, M, N)
TetX = np.rad2deg(np.arcsin(-M))
TetY = np.rad2deg(np.arcsin(L / np.cos(np.arcsin(-M))))
print(TetX, TetY)
print("---------------------------------------------------------------")

# ______________________________________#

Pup.Samp = 10
Pup.Ptype = "hexapolar"
Pup.FieldY = 0.0
Pup.FieldType = "angle"
x, y, z, L, M, N = Pup.Pattern2Field()
Rayos = Kos.raykeeper(Telescopio)

# ______________________________________#

for i in range(0, len(x)):
    pSource_0 = [x[i], y[i], z[i]]
    dCos = [L[i], M[i], N[i]]
    W = 0.4
    Telescopio.Trace(pSource_0, dCos, W)
    Rayos.push()

# ______________________________________#

Kos.display2d(Telescopio, Rayos, 1, 1)

# ______________________________________#

X, Y, Z, L, M, N = Rayos.pick(-1)
plt.figure(300)
plt.plot(X, Y, 'x')
plt.axis('square')
plt.show(block=False)




import numpy as np
import math
from .SurfTools import surface_tools as SUT



class InterNormalCalc():
    """InterNormalCalc.
    """


    def __init__(self, SurfData, TypTot, PR3D, HS):
        """__init__.

        Parameters
        ----------
        SurfData :
            SurfData
        TypTot :
            TypTot
        PR3D :
            PR3D
        HS :
            HS
        """
        self.HS = HS
        self.SDT = SurfData
        self.n = len(self.SDT)
        self.SuTo = SUT(SurfData)
        self.Pr3D = PR3D
        self.Disable_Inner = 1
        self.ExtraDiameter = 0
        self.AAA = self.Pr3D.AAA
        self.BBB = self.Pr3D.BBB
        self.DDD = self.Pr3D.DDD
        self.EEE = self.Pr3D.EEE
        self.GlassOnSide = self.Pr3D.GlassOnSide
        self.side_number = self.Pr3D.side_number
        self.TypeTotal = self.Pr3D.TypeTotal
        self.TRANS_1A = self.Pr3D.TRANS_1A
        self.TRANS_2A = self.Pr3D.TRANS_2A
        self.Pn = np.asarray([0.0, 0.0, 0.0])

        self.P1 = np.asarray([0.0, 0.0, 0.0, 1.0])
        self.P2 = np.asarray([0.0, 0.0, 0.0, 1.0])
        self.P_z1 = 10000000.0

    def __SigmaHitTransfSpace(self, PP_start, PP_stop, j):
        """__SigmaHitTransfSpace.

        Parameters
        ----------
        PP_start :
            PP_start
        PP_stop :
            PP_stop
        j :
            j
        """
        # if (len(np.shape(PP_start)) != 1):
        #     PP_stop_zyx = np.rot90(PP_stop)
        #     PP_start_zyx = np.rot90(PP_start)
        #     StopPoint = np.array([PP_stop_zyx[2], PP_stop_zyx[1], PP_stop_zyx[0], 1.0])
        #     StarPoint = np.array([PP_start_zyx[2], PP_start_zyx[1], PP_start_zyx[0], 1.0])
        # else:

        StopPoint = np.array([PP_stop[0], PP_stop[1], PP_stop[2], 1.0])
        StarPoint = np.array([PP_start[0], PP_start[1], PP_start[2], 1.0])

        SurfHit = 1
        P_SurfHit = self.Pr3D.TRANS_1A[j].dot(StopPoint)
        Px1 = P_SurfHit[(0, 0)]
        Py1 = P_SurfHit[(0, 1)]
        Pz1 = P_SurfHit[(0, 2)]

        P_start = self.Pr3D.TRANS_1A[j].dot(StarPoint)
        P_x1 = P_start[(0, 0)]
        P_y1 = P_start[(0, 1)]
        P_z1 = P_start[(0, 2)]

        P12 = [(Px1 - P_x1), (Py1 - P_y1), (Pz1 - P_z1)]
        [L, M, N] = (P12 / np.linalg.norm(P12))

        Px1 = (((L / N) * (- P_z1)) + P_x1)
        Py1 = (((M / N) * (- P_z1)) + P_y1)

        Pz1 = 0
        SurfHit = 1

        P_x2 = 0
        P_y2 = 0
        P_z2 = 0

        if (len(np.shape(P_x1)) != 0):
            L = L[0]
            M = M[0]
            N = N[0]
            P_x1 = P_x1[0]
            P_y1 = P_y1[0]
            P_z1 = P_z1[0]
            Px1 = Px1[0]
            Py1 = Py1[0]
            Pz1 = Pz1[0]
            P_x2 = P_x2[0]
            P_y2 = P_y2[0]
            P_z2 = P_z2[0]
            SurfHit = SurfHit[0]

        if (self.SDT[j].Thin_Lens == 0):
            self.vj = j

            ASD=np.sqrt(((Px1-self.SDT[j].SubAperture[2])**2) + ((Py1-self.SDT[j].SubAperture[1])**2))
            D0 = (2.0 * ASD)
            DiamInf = ((self.SDT[j].InDiameter * self.SDT[j].SubAperture[0]) * self.Disable_Inner)
            DiamSup = ((self.SDT[j].Diameter * self.SDT[j].SubAperture[0]) + (10000.0 * self.ExtraDiameter))
            if ((D0 > DiamSup) or (D0 < DiamInf)):
                SurfHit = 0
                P_x2 = 0
                P_y2 = 0
                P_z2 = 0
            else:
                (P_x2, P_y2, P_z2) = self.HS.SolveHit(Px1, Py1, Pz1, L, M, N, j)
                if (not math.isnan(P_z2)):
                    P_x2 = self.HS.vevaX
                    P_y2 = self.HS.vevaY

                    ASD=np.sqrt(((Px1-self.SDT[j].SubAperture[2])**2) + ((Py1-self.SDT[j].SubAperture[1])**2))
                    D0 = (2.0 * ASD)

                    DiamInf = ((self.SDT[j].InDiameter * self.SDT[j].SubAperture[0]) * self.Disable_Inner)
                    DiamSup = ((self.SDT[j].Diameter * self.SDT[j].SubAperture[0]) + (10000.0 * self.ExtraDiameter))
                    if ((D0 > DiamSup) or (D0 < DiamInf)):
                        SurfHit = 0
                else:
                    SurfHit = 0
                    P_x2 = 0
                    P_y2 = 0
                    P_z2 = 0
        else:
            ASD=np.sqrt(((Px1 - self.SDT[j].SubAperture[2])**2) + ((Py1 - self.SDT[j].SubAperture[1])**2))
            D0 = (2.0 * ASD)*0.999999
            #0.9999 to avoid kill the rays in the edge

            if ((D0 > self.SDT[j].Diameter * self.SDT[j].SubAperture[0]) or (D0 < self.SDT[j].InDiameter * self.SDT[j].SubAperture[0] )):
                SurfHit = 0
                P_x2 = 0
                P_y2 = 0
                P_z2 = 0

            else:
                P_x2 = ((L / N) * self.SDT[j].Thin_Lens)
                P_y2 = ((M / N) * self.SDT[j].Thin_Lens)
                P_z2 = self.SDT[j].Thin_Lens

        return (SurfHit, P_x2, P_y2, P_z2, Px1, Py1, Pz1, L, M, N)

    def __SigmaHitTransfSpaceFast(self, PP_start, PP_stop, j):

        StopPoint = np.array([PP_stop[0], PP_stop[1], PP_stop[2], 1.0])
        StarPoint = np.array([PP_start[0], PP_start[1], PP_start[2], 1.0])

        SurfHit = 1
        P_SurfHit = self.Pr3D.TRANS_1A[j].dot(StopPoint)
        Px1 = P_SurfHit[(0, 0)]
        Py1 = P_SurfHit[(0, 1)]
        Pz1 = P_SurfHit[(0, 2)]

        P_start = self.Pr3D.TRANS_1A[j].dot(StarPoint)
        P_x1 = P_start[(0, 0)]
        P_y1 = P_start[(0, 1)]
        P_z1 = P_start[(0, 2)]

        P12 = [(Px1 - P_x1), (Py1 - P_y1), (Pz1 - P_z1)]
        [L, M, N] = (P12 / np.linalg.norm(P12))

        Px1 = (((L / N) * (- P_z1)) + P_x1)
        Py1 = (((M / N) * (- P_z1)) + P_y1)

        Pz1 = 0
        SurfHit = 1

        P_x2 = 0
        P_y2 = 0
        P_z2 = 0


        (P_x2, P_y2, P_z2) = self.HS.SolveHit(Px1, Py1, Pz1, L, M, N, j)
        if (not math.isnan(P_z2)):
            P_x2 = self.HS.vevaX
            P_y2 = self.HS.vevaY

        else:
            SurfHit = 0
            P_x2 = 0
            P_y2 = 0
            P_z2 = 0


        return (SurfHit, P_x2, P_y2, P_z2, Px1, Py1, Pz1)

    def __ParaxCalcObjOut2OrigSpace(self, Px2, Py2, Pz2, Px1, Py1, Pz1, j):
        """__ParaxCalcObjOut2OrigSpace.

        Parameters
        ----------
        Px2 :
            Px2
        Py2 :
            Py2
        Pz2 :
            Pz2
        Px1 :
            Px1
        Py1 :
            Py1
        Pz1 :
            Pz1
        j :
            j
        """
        P1 = [Px1, Py1, Pz1, 1]
        P2 = [Px2, Py2, Pz2, 1]
        NP1 = self.TRANS_2A[j].dot(P1)
        NP2 = self.TRANS_2A[j].dot(P2)
        Pn = np.asarray([(- (NP1[(0, 0)] - NP2[(0, 0)])), (- (NP1[(0, 1)] - NP2[(0, 1)])), (- (NP1[(0, 2)] - NP2[(0, 2)]))])
        norm = (Pn / np.linalg.norm(Pn))
        PTO_exit = [NP1[(0, 0)], NP1[(0, 1)], NP1[(0, 2)]]
        PTO_exit_Object_Space = [Px1, Py1, Pz1]
        return (norm, PTO_exit, PTO_exit_Object_Space)

    def __SigmaOutOrigSpace(self, P_x2, P_y2, P_z2, j):
        """__SigmaOutOrigSpace.

        Parameters
        ----------
        P_x2 :
            P_x2
        P_y2 :
            P_y2
        P_z2 :
            P_z2
        j :
            j
        """


        (New_L, New_M, New_N) = self.HS.SurfDer(P_x2, P_y2, P_z2)


        Pz1z2 = (self.P_z1 - P_z2)

        P_x1 = ((Pz1z2 * (New_L / New_N)) + P_x2)
        P_y1 = ((Pz1z2 * (New_M / New_N)) + P_y2)


        self.P1[0], self.P1[1], self.P1[2] = P_x1, P_y1, self.P_z1
        self.P2[0], self.P2[1], self.P2[2] = P_x2, P_y2, P_z2

        NP1 = self.TRANS_2A[j].dot(self.P1)
        NP2 = self.TRANS_2A[j].dot(self.P2)


        self.Pn[0] = - (NP1[(0, 0)] - NP2[(0, 0)])
        self.Pn[1] = - (NP1[(0, 1)] - NP2[(0, 1)])
        self.Pn[2] = - (NP1[(0, 2)] - NP2[(0, 2)])


        LNOR=np.sqrt((self.Pn[0]**2.)+(self.Pn[1]**2.)+(self.Pn[2]**2.))

        norm = (self.Pn / LNOR)

        PTO_exit = [NP2[(0, 0)], NP2[(0, 1)], NP2[(0, 2)]]
        PTO_exit_Object_Space = [P_x2, P_y2, P_z2]


        return (norm, PTO_exit, PTO_exit_Object_Space)

    def __HitOnMask(self, PP_start, PP_stop, j):
        """__HitOnMask.

        Parameters
        ----------
        PP_start :
            PP_start
        PP_stop :
            PP_stop
        j :
            j
        """
        SurfHit = 1
        HITS_CONT = []
        if (self.SDT[j].Mask_Type != 0):
            OBJECT = self.DDD[j]
            for obj in OBJECT:
                (inter_mask, ind_mask) = obj.ray_trace(PP_start, PP_stop)
                Hit_MASK = np.shape(inter_mask)[0]
                HITS_CONT.append(Hit_MASK)
            HITS_CONT = np.asarray(HITS_CONT)
            if np.any((HITS_CONT == 1)):
                SurfHit_MASK = 1
            else:
                SurfHit_MASK = 0
            if (self.SDT[j].Mask_Type == 1):
                if (SurfHit_MASK == 1):
                    SurfHit = 1
                else:
                    SurfHit = 0
            if (self.SDT[j].Mask_Type == 2):
                if (SurfHit_MASK == 1):
                    SurfHit = 0
                else:
                    SurfHit = 1
        return SurfHit

    def __GrooveDirectionVector(self, j):
        """__GrooveDirectionVector.

        Parameters
        ----------
        j :
            j
        """



        # self.P1[0], self.P1[1], self.P1[2] = 0, 0, 0
        # self.P2[0], self.P2[1], self.P2[2] = np.sin(np.deg2rad(self.SDT[j].Grating_Angle)), np.cos(np.deg2rad(self.SDT[j].Grating_Angle)), 0

        self.P1[0], self.P1[1], self.P1[2] = 0, 0, 0
        self.P2[0], self.P2[1], self.P2[2] = -np.cos(np.deg2rad(self.SDT[j].Grating_Angle)), -np.sin(np.deg2rad(self.SDT[j].Grating_Angle)),0

        NP1 = self.TRANS_2A[j].dot(self.P1)
        NP2 = self.TRANS_2A[j].dot(self.P2)


        self.Pn[0] = - (NP1[(0, 0)] - NP2[(0, 0)])
        self.Pn[1] = - (NP1[(0, 1)] - NP2[(0, 1)])
        self.Pn[2] = - (NP1[(0, 2)] - NP2[(0, 2)])


        LNOR=np.sqrt((self.Pn[0]**2.)+(self.Pn[1]**2.)+(self.Pn[2]**2.))

        Pg_v = (self.Pn / LNOR)





        return Pg_v

    def InterNormal(self, PP_start, PP_stop, j, jj):
        """InterNormal.

        Parameters
        ----------
        PP_start :
            PP_start
        PP_stop :
            PP_stop
        j :
            j
        jj :
            jj
        """
        PTO_exit = [0, 0, 0]
        PTO_exit_Object_Space = [0, 0, 0]
        LMN_exit_Object_Space = [0, 0, 1]
        norm = [0, 0, 1]
        SurfHit = 1

        if (self.SDT[j].Diff_Ord == 0):
            Pgn = [0, 1, 0]

        else:
            Pgn = self.__GrooveDirectionVector(j)

        if (self.TypeTotal[jj] == 0):
            SurfHit = 1
            SurfHit = self.__HitOnMask(PP_start, PP_stop, j)

            if (SurfHit != 0):

                (SurfHit, Px2, Py2, Pz2, Px1, Py1, Pz1, L, M, N) = self.__SigmaHitTransfSpace(PP_start, PP_stop, j)
                LMN_exit_Object_Space = [L, M, N]

                if (self.SDT[j].Thin_Lens == 0):



                    (norm, PTO_exit, PTO_exit_Object_Space) = self.__SigmaOutOrigSpace(Px2, Py2, Pz2, j)

                else:
                    (norm, PTO_exit, PTO_exit_Object_Space) = self.__ParaxCalcObjOut2OrigSpace(Px2, Py2, Pz2, Px1, Py1, Pz1, j)
        else:
            (SurfHit, norm, PTO_exit, Pgn) = self.__InterNormalSolidObject(jj, PP_start, PP_stop)
        return (SurfHit, np.asarray(norm), np.asarray(PTO_exit), np.asarray(Pgn), np.asarray(PTO_exit_Object_Space), np.asarray(LMN_exit_Object_Space), j)



    def InterNormalFast(self, PP_start, PP_stop, j):
# Sin solid objects, sin mascaras ni paraxial, no tilts, no inner or out
        """InterNormal.

        Parameters
        ----------
        PP_start :
            PP_start
        PP_stop :
            PP_stop
        j :
            j
        jj :
            jj
        """
        PTO_exit = [0, 0, 0]
        PTO_exit_Object_Space = [0, 0, 0]
        norm = [0, 0, 1]
        SurfHit = 1

        if (self.SDT[j].Diff_Ord == 0):
            Pgn = [0, 1, 0]
        else:
            Pgn = self.__GrooveDirectionVector(j)

        (SurfHit, Px2, Py2, Pz2, Px1, Py1, Pz1) = self.__SigmaHitTransfSpaceFast(PP_start, PP_stop, j)

        if (SurfHit != 0):
            (norm, PTO_exit, PTO_exit_Object_Space) = self.__SigmaOutOrigSpace(Px2, Py2, Pz2, j)

        return (SurfHit, np.asarray(norm), np.asarray(PTO_exit), np.asarray(Pgn), np.asarray(PTO_exit_Object_Space), j)



    def __InterNormalSolidObject(self, jj, PP_start, PP_stop):
        """__InterNormalSolidObject.

        Parameters
        ----------
        jj :
            jj
        PP_start :
            PP_start
        PP_stop :
            PP_stop
        """
        Pgn = np.asarray([0, 1, 0])
        PTO_exit = [0, 0, 0]
        norm = [0, 0, 1]
        (inter, ind) = self.EEE[jj].ray_trace(PP_start, PP_stop)
        SurfHit = np.shape(inter)[0]
        if (SurfHit != 0):
            s = 0
            h = []
            for f in ind:
                PD = (np.asarray(inter[s]) - np.asarray(PP_start))
                distance = np.linalg.norm(PD)
                if (np.abs(distance) < 0.05):
                    distance = 99999999999999.9
                h.append(distance)
                s = (s + 1)
            index = np.argmin(np.asarray(h))
            PTO_exit = inter[index]
            NOR = self.EEE[jj].cell_normals
            norm = NOR[ind[index]]
            Pgn = np.asarray([0, 0, 1])
        return (SurfHit, norm, PTO_exit, Pgn)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Tel 2M Spyder Spot Diagram"""

# import os
import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Thickness = 2.000000000000000E+003
P_Obj.Glass = "AIR"
P_Obj.Diameter = 6.796727741707513E+002 * 2.0
P_Obj.Drawing = 0

# ______________________________________#

M1 = Kos.surf()
M1.Rc = -6.0E+003
M1.Thickness = -1.774190000000000E+003 + 1.853722901194000E+000
M1.k = -1.6+000
M1.Glass = "MIRROR"
M1.Diameter = 6.63448E+002 * 2.0
M1.InDiameter = 228.6 * 2.0
M1.DespY = 0.0
M1.TiltX = 0.0000
M1.AxisMove = 1

# ______________________________________#

M2 = Kos.surf()
M2.Rc = -6.00E+003
M2.Thickness = -M1.Thickness
M2.k = -3.4782E+001
M2.Glass = "MIRROR"
M2.Diameter = 3.0E+002 * 2.0

# ______________________________________#

Vertex = Kos.surf()
Vertex.Thickness = 130.0
Vertex.Glass = "AIR"
Vertex.Diameter = 600.0
Vertex.Drawing = 0

# ______________________________________#

file = r"Prisma.stl"

# ______________________________________#

objeto = Kos.surf()
objeto.Diameter = 118.0 * 2.0
objeto.Solid_3d_stl = file
objeto.Thickness = 600
objeto.Glass = "BK7"
objeto.TiltX = 55
objeto.TiltY = 0
objeto.TiltZ = 45
objeto.DespX = 0
objeto.DespY = 0
objeto.AxisMove = 0

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Rc = 0
P_Ima.Thickness = 100.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 500.0
P_Ima.Drawing = 1

# ______________________________________#

A = [P_Obj, M1, M2, Vertex, objeto, P_Ima]
configuracion_1 = Kos.Setup()

# ______________________________________#

Telescope = Kos.system(A, configuracion_1)
Rays = Kos.raykeeper(Telescope)




Telescope.energy_probability = 1
Telescope.NsLimit = 10

# ______________________________________#

W = 0.633
tam = 5
rad = 6.56727741707513E+002
tsis = len(A) + 2
for gg in range(0, 10):
    for j in range(-tam, tam + 1):
        # j=0
        for i in range(-tam, tam + 1):
            x_0 = (i / tam) * rad
            y_0 = (j / tam) * rad
            r = np.sqrt((x_0 * x_0) + (y_0 * y_0))
            if r < rad:
                tet = 0.0
                pSource_0 = [x_0, y_0, 0.0]
                # print("-...............")
                dCos = [0.0, np.sin(np.deg2rad(tet)), np.cos(np.deg2rad(tet))]
                W = 0.633
                Telescope.NsTrace(pSource_0, dCos, W)
                Rays.push()

# # ______________________________________#

Kos.display3d(Telescope, Rays, 0)
print(Telescope.EFFL)



from setuptools import setup

setup(
    name='KrakenOS',
    version='1.0.0.16',
    packages=['KrakenOS', 'KrakenOS.AstroAtmosphere', 'KrakenOS.Cat', 'KrakenOS.Examples', 'KrakenOS.Docs'],

    install_requires=['pyvista','PyVTK','vtk','numpy','scipy','matplotlib', 'csv342', 'numba' ],
    package_data={'': ['LICENSE.txt', '../*.AGF', '../*.agf', '../*.stl', '../*.scad', '../*.pdf']},
    include_package_data=True,
    url='https://github.com/Garchupiter/Kraken-Optical-Simulator',
    download_url='https://github.com/Garchupiter/Kraken-Optical-Simulator',
    keywords=['Optical', 'Simulator', 'lens'],
    license='GNU General Public License v3.0',
    author='joel Herrera et al.',
    author_email='joel@astro.unam.mx',
    description='Optical Simulation and ray tracing '
)




from scipy.stats import uniform
import random
import numpy as np

class SourceRnd():
    """SourceRnd.

        SourceRnd.fun = Python function of angle to cover
        SourceRnd.field = Angle to cover from pole (Deg)
        SourceRnd.type = 0 for circle 1 to square
        SourceRnd.dim = Lateral size
        SourceRnd.num = Number of rays

    """


    def __init__(self):
        """__init__.
        """
        self.fun = 0
        self.field = 89.9
        self.type = 0
        self.dim = 10
        self.num = 100

    def rays(self):
        """rays.
        """

        def ff(x):
            """ff.

            Parameters
            ----------
            x :
                x
            """
            return (x / x)
        f = self.fun
        if (f == 0):
            f = ff
        r_source = self.dim
        lim = np.deg2rad(self.field)
        A = np.linspace((lim / self.num), (lim - (lim / self.num)), self.num)
        prob = (f(A) * ((2 * np.pi) * np.sin(A)))
        prob = (prob - np.min(prob))
        prob = (prob / np.max(prob))
        Rand_num = random.choices(A, prob, k=self.num)
        Delta = ((lim / self.num) * (uniform.rvs(size=self.num, loc=0, scale=2) - 1))
        Theta = (Rand_num + Delta)
        Phi = uniform.rvs(size=self.num, loc=0, scale=(2.0 * np.pi))
        L = (np.sin(Theta) * np.cos(Phi))
        M = (np.sin(Theta) * np.sin(Phi))
        N = np.cos(Theta)
        if (self.type == 0):
            AR = np.linspace((r_source / self.num), (r_source - (r_source / self.num)), self.num)
            prob_R = ((2 * np.pi) * AR)
            lt0 = np.argwhere((prob_R < 0))
            prob_R = (prob_R - np.min(prob_R))
            prob_R = (prob_R / np.max(prob_R))
            Rand_num_R = random.choices(AR, prob_R, k=self.num)
            Delta_R = ((lim / self.num) * (uniform.rvs(size=self.num, loc=0, scale=2) - 1))
            R = (Rand_num_R + Delta_R)
            Phi_R = uniform.rvs(size=self.num, loc=0, scale=(2.0 * np.pi))
            X = (R * np.cos(Phi_R))
            Y = (R * np.sin(Phi_R))
        if (self.type == 1):
            X = (r_source * (uniform.rvs(size=self.num, loc=0, scale=2) - 1))
            Y = (r_source * (uniform.rvs(size=self.num, loc=0, scale=2) - 1))
        Z = np.zeros_like(X)
        return (L, M, N, X, Y, Z)





import numpy as np

class surface_tools():
    """surface_tools.
    """


    def __init__(self, SurfData):
        """__init__.

        Parameters
        ----------
        SurfData :
            SurfData
        """
        self.SDT = SurfData
        self.ErrSurfCase = 0
        self.Surface_Flattener = (- 1)

    def SurfaceShape(self, x, y, j):
        """SurfaceShape.

        Parameters
        ----------
        x :
            x
        y :
            y
        j :
            j
        """
        if (j != self.Surface_Flattener):
            # self.Surface_Flattener
            TOTAL_SURF_SHAPE = self.SDT[j].sigma_z(x, y, self.ErrSurfCase)
        else:
            TOTAL_SURF_SHAPE = 0.0*np.copy(x)
        return TOTAL_SURF_SHAPE




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Doublet Lens Tilt"""

import numpy as np
import pkg_resources
""" Looking for if KrakenOS is installed, if not, it assumes that
an folder downloaded from github is run"""

required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("Not installed")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0.0
P_Obj.Thickness = 10
P_Obj.Glass = "AIR"
P_Obj.Diameter = 30.0

# ______________________________________#

L1a = Kos.surf()
L1a.Rc = 9.284706570002484E+001
L1a.Thickness = 6.0
L1a.Glass = "BK7"
L1a.Diameter = 30.0
L1a.AxisMove = 1
L1a.TiltX = 1
L1a.DespY = 10.

# ______________________________________#

L1b = Kos.surf()
L1b.Rc = (-3.071608670000159E+001)
L1b.Thickness = 3.0
L1b.Glass = "F2"
L1b.Diameter = 30

# ______________________________________#

L1c = Kos.surf()
L1c.Rc = (-7.819730726078505E+001)
L1c.Thickness = 9.737604742910693E+001
L1c.Glass = "AIR"
L1c.Diameter = 30

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Rc = 0.0
P_Ima.Thickness = 0.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 100.0
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, L1a, L1b, L1c, P_Obj, L1a, L1b, L1c, P_Ima]
configuracion_1 = Kos.Setup()

# ______________________________________#

Doblete = Kos.system(A, configuracion_1)
Rayos = Kos.raykeeper(Doblete)

# ______________________________________#

tam = 5
rad = 10.0
tsis = len(A) - 1
for i in range(-tam, tam + 1):
    for j in range(-tam, tam + 1):
        x_0 = (i / tam) * rad
        y_0 = (j / tam) * rad
        r = np.sqrt((x_0 * x_0) + (y_0 * y_0))
        if r < rad:
            tet = 0.0
            pSource_0 = [x_0, y_0, 0.0]
            dCos = [0.0, np.sin(np.deg2rad(tet)), np.cos(np.deg2rad(tet))]
            W = 0.4
            Doblete.Trace(pSource_0, dCos, W)
            Rayos.push()
            W = 0.5
            Doblete.Trace(pSource_0, dCos, W)
            Rayos.push()
            W = 0.6
            Doblete.Trace(pSource_0, dCos, W)
            Rayos.push()

# ______________________________________#

Kos.display3d(Doblete, Rayos, 2)
Kos.display2d(Doblete, Rayos, 0)




import numpy as np
from .RayKeeper import *
from .Display import *
import scipy
from scipy import optimize
from scipy.optimize import fmin_cg
from scipy.optimize import fsolve
from .AstroAtmosphere import *
""" v2. """

def RMS_Pupil(r, SYSTEM, Surf, W, tet):
    """RMS_Pupil.

    Parameters
    ----------
    r :
        r
    SYSTEM :
        SYSTEM
    Surf :
        Surf
    W :
        W
    """
    SYSTEM.TargSurf(Surf)
    SYSTEM.IgnoreVignetting(0)
    SYSTEM.SurfFlat(Surf, 0)
    RP = raykeeper(SYSTEM)
    # tet = 0.01
    s_0 = [0.0, 0.0, 0.0]
    c_0 = [0.0, 0.0, 1.0]
    SYSTEM.Trace(s_0, c_0, W)
    RP.push()

    c_1 = [np.sin(np.deg2rad(tet)), 0.0, np.cos(np.deg2rad(tet))]
    s_0 = [r, 0.0, 0.0]
    SYSTEM.Trace(s_0, c_1, W)
    RP.push()

    c_2 = [np.sin(np.deg2rad((- tet))), 0.0, np.cos(np.deg2rad((- tet)))]
    s_0 = [(- r), 0.0, 0.0]
    SYSTEM.Trace(s_0, c_2, W)
    RP.push()

    c_3 = [0.0, np.sin(np.deg2rad(tet)), np.cos(np.deg2rad(tet))]
    s_0 = [0.0, r, 0.0]
    SYSTEM.Trace(s_0, c_3, W)
    RP.push()

    c_4 = [0.0, np.sin(np.deg2rad((- tet))), np.cos(np.deg2rad((- tet)))]
    s_0 = [0.0, (- r), 0.0]
    SYSTEM.Trace(s_0, c_4, W)
    RP.push()

    (X, Y, Z, L, M, N) = RP.pick(Surf)
    delta_Z = 0
    X = (((L / N) * delta_Z) + X)
    Y = (((M / N) * delta_Z) + Y)
    cenX = np.mean(X)
    cenY = np.mean(Y)
    x1 = (X - cenX)
    y1 = (Y - cenY)
    R2 = ((x1 * x1) + (y1 * y1))
    R_RMS = np.sqrt(np.mean(R2))
    SYSTEM.SurfFlat((- 1))
    SYSTEM.TargSurf((- 1))
    SYSTEM.Vignetting(0)
    RP.clean()
    return R_RMS

def R_RMS(delta_Z, L, M, N, X, Y):
    """R_RMS.

    Parameters
    ----------
    delta_Z :
        delta_Z
    L :
        L
    M :
        M
    N :
        N
    X :
        X
    Y :
        Y
    """
    X = (((L / N) * delta_Z) + X)
    Y = (((M / N) * delta_Z) + Y)
    cenX = np.mean(X)
    cenY = np.mean(Y)
    x1 = (X - cenX)
    y1 = (Y - cenY)
    R2 = ((x1 * x1) + (y1 * y1))
    R_RMS = np.sqrt(np.mean(R2))
    return R_RMS

def FuncVectorCross(Z1, XYZa, LMNa, XYZb, LMNb, xy):
    """FuncVectorCross.

    Parameters
    ----------
    Z1 :
        Z1
    XYZa :
        XYZa
    LMNa :
        LMNa
    XYZb :
        XYZb
    LMNb :
        LMNb
    xy :
        xy
    """
    [La, Ma, Na] = LMNa
    [X0a, Y0a, Z0a] = XYZa
    [Lb, Mb, Nb] = LMNb
    [X0b, Y0b, Z0b] = XYZb
    if (xy == 0):
        V = ((((La / Na) * (Z1 - Z0a)) + X0a) - (((Lb / Nb) * (Z1 - Z0b)) + X0b))
    else:
        V = ((((Ma / Na) * (Z1 - Z0a)) + Y0a) - (((Mb / Nb) * (Z1 - Z0b)) + Y0b))
    return V

def DerVectCross(Z1, XYZa, LMNa, XYZb, LMNb, xy):
    """DerVectCross.

    Parameters
    ----------
    Z1 :
        Z1
    XYZa :
        XYZa
    LMNa :
        LMNa
    XYZb :
        XYZb
    LMNb :
        LMNb
    xy :
        xy
    """
    h = 1e-07
    f1 = FuncVectorCross((Z1 + h), XYZa, LMNa, XYZb, LMNb, xy)
    f2 = FuncVectorCross((Z1 - h), XYZa, LMNa, XYZb, LMNb, xy)
    der = ((f1 - f2) / (2 * h))
    return der

def SolveVectCross(XYZa, LMNa, XYZb, LMNb, xy):
    """SolveVectCross.

    Parameters
    ----------
    XYZa :
        XYZa
    LMNa :
        LMNa
    XYZb :
        XYZb
    LMNb :
        LMNb
    xy :
        xy
    """
    [La, Ma, Na] = LMNa
    [X0a, Y0a, Z0a] = XYZa
    [Lb, Mb, Nb] = LMNb
    [X0b, Y0b, Z0b] = XYZb
    Z1 = 1e-07
    cnt = 0
    while True:
        fun = FuncVectorCross(Z1, XYZa, LMNa, XYZb, LMNb, xy)
        derfun = DerVectCross(Z1, XYZa, LMNa, XYZb, LMNb, xy)
        Z2 = (Z1 - (fun / derfun))
        if (np.abs((Z1 - Z2)) < 1e-06):
            break
        else:
            Z1 = Z2
        if (cnt == 5):
            break
        cnt = (cnt + 1)
    X2 = (((La / Na) * (Z2 - Z0a)) + X0a)
    Y2 = (((Ma / Na) * (Z2 - Z0a)) + Y0a)
    return [X2, Y2, Z2]

def DerFpupil(SYSTEM, XY, H, Surf, tet, W, xy):
    """DerFpupil.

    Parameters
    ----------
    SYSTEM :
        SYSTEM
    XY :
        XY
    H :
        H
    Surf :
        Surf
    tet :
        tet
    W :
        W
    xy :
        xy
    """
    h = 1e-07
    der = (Fpupil(SYSTEM, (XY + h), H, Surf, tet, W, xy) - Fpupil(SYSTEM, (XY - h), H, Surf, tet, W, xy))
    DER = (der / (2.0 * h))
    return DER

def Fpupil(SYSTEM, XY, H, Surf, tet, W, xy):
    """Fpupil.

    Parameters
    ----------
    SYSTEM :
        SYSTEM
    XY :
        XY
    H :
        H
    Surf :
        Surf
    tet :
        tet
    W :
        W
    xy :
        xy
    """
    if (xy == 0):
        pSource_0 = [XY, 0.0, 0.0]
        dCos = [np.sin(np.deg2rad(tet)), 0.0, np.cos(np.deg2rad(tet))]
    else:
        pSource_0 = [0.0, XY, 0.0]
        dCos = [0.0, np.sin(np.deg2rad(tet)), np.cos(np.deg2rad(tet))]
    SYSTEM.Trace(pSource_0, dCos, W)
    s = np.asarray(SYSTEM.SurfACE)
    a = np.squeeze(np.argwhere((s == Surf)))
    [X2, Y2, Z2] = SYSTEM.OST_XYZ[a]
    if (xy == 0):
        R = (X2 - H)
    else:
        R = (Y2 - H)
    return R

def SolveRayPupil(SYSTEM, H, tet, W, Surf, xy):
    """SolveRayPupil.

    Parameters
    ----------
    SYSTEM :
        SYSTEM
    H :
        H
    tet :
        tet
    W :
        W
    Surf :
        Surf
    xy :
        xy
    """
    XY0 = 1e-07
    cnt = 0
    while True:
        XY1 = (XY0 - (Fpupil(SYSTEM, XY0, H, Surf, tet, W, xy) / DerFpupil(SYSTEM, XY0, H, Surf, tet, W, xy)))
        if (np.abs((XY0 - XY1)) < 1e-06):
            break
        else:
            XY0 = XY1
        if (cnt == 5):
            break
        cnt = (cnt + 1)
    if (xy == 0):
        pSource_0 = [XY0, 0.0, 0.0]
        dCos = [np.sin(np.deg2rad(tet)), 0.0, np.cos(np.deg2rad(tet))]
    else:
        pSource_0 = [0.0, XY0, 0.0]
        dCos = [0.0, np.sin(np.deg2rad(tet)), np.cos(np.deg2rad(tet))]
    return (pSource_0, dCos)

class PupilCalc():
    """PupilCalc.
    """


    def __init__(self, system, Surf, W, ApTyp='STOP', AV=1.0):
        """__init__.

        Parameters
        ----------
        system :
            system
        Surf :
            Surf
        W :
            W
        ApTyp :
            ApTyp
        AV :
            AV
        """
        self.Surf = Surf
        self.W = W
        self.SYSTEM = system
        if (AV == 0):
            AV = 1.0
        self.FieldType = 'angle'
        self.ApertureType = ApTyp
        self.ApertureValue = AV
        self.x0 = 0
        self.y0 = 0
        self.z0 = 0
        self.L = 0.0
        self.M = 0.0
        self.N = 1.0
        self.Cordx = np.asarray(0)
        self.Cordy = np.asarray(0)
        self.Ptype = 'hexapolar'
        self.Samp = 6
        self.FieldX = 0.0
        self.FieldY = 0.0
        self.DirPupSal = [0.0, 0.0, 0.0]
        self.rad = 0
        self.theta = 0
        self.PupilInpFactor = 0.99
        self.menter = 1.0
        self.AtmosRef = 0
        self.T = 283.15
        self.P = 100500
        self.H = 0.0
        self.xc = 450
        self.lat = 50
        self.h = 0
        self.l1 = 5000.60169
        self.l2 = 0.50169
        self.z0 = 75.0
        self.tet= 0.01

        delta_Z = 0
        r = 1e-06
        cnt = 0
        while True:
            fun = RMS_Pupil(r, self.SYSTEM, Surf, self.W, self.tet)
            h = 1e-7
            f1 = RMS_Pupil((r + h), self.SYSTEM, Surf, self.W, self.tet)
            f2 = RMS_Pupil((r - h), self.SYSTEM, Surf, self.W, self.tet)
            der = ((f1 - f2) / (2 * h))
            r2 = (r - (fun / der))
            if (np.abs((r - r2)) < 1e-7):
                break
            else:
                r = r2
            if (cnt == 5):
                break
            cnt = (cnt + 1)
        self.SYSTEM.IgnoreVignetting(0)
        RP = raykeeper(self.SYSTEM)


        s_0 = [0.0, 0.0, 0.0]
        c_0 = [0.0, 0.0, 1.0]
        self.SYSTEM.Trace(s_0, c_0, self.W)

        RP.push()
        c_1 = [np.sin(np.deg2rad(self.tet)), 0.0, np.cos(np.deg2rad(self.tet))]
        s_0 = [r, 0.0, 0.0]
        self.SYSTEM.Trace(s_0, c_1, self.W)
        RP.push()

        c_2 = [np.sin(np.deg2rad((- self.tet))), 0.0, np.cos(np.deg2rad((- self.tet)))]
        s_0 = [(- r), 0.0, 0.0]
        self.SYSTEM.Trace(s_0, c_2, self.W)
        RP.push()

        c_3 = [0.0, np.sin(np.deg2rad(self.tet)), np.cos(np.deg2rad(self.tet))]
        s_0 = [0.0, r, 0.0]
        self.SYSTEM.Trace(s_0, c_3, self.W)
        RP.push()

        c_4 = [0.0, np.sin(np.deg2rad((- self.tet))), np.cos(np.deg2rad((- self.tet)))]
        s_0 = [0.0, (- r), 0.0]
        self.SYSTEM.Trace(s_0, c_4, self.W)
        RP.push()

# ----------------------------------------------------------------------------

        (X, Y, Z, L, M, N) = RP.pick(0)
        theta = 0

        for s in range(1, 5):
            theta = (theta + np.rad2deg(np.arccos((((L[0] * L[s]) + (M[0] * M[s])) + (N[0] * N[s])))))
        thetaIni = np.abs((theta / 4.0))

# ----------------------------------------------------------------------------

        (X, Y, Z, L, M, N) = RP.pick(Surf)
        theta = 0

        for s in range(1, 5):
            theta = (theta + np.rad2deg(np.arccos((((L[0] * L[s]) + (M[0] * M[s])) + (N[0] * N[s])))))
        thetaSurf = np.abs((theta / 4.0))

# ----------------------------------------------------------------------------

        (X, Y, Z, L, M, N) = RP.pick((- 1))
        theta = 0

        for s in range(1, 5):
            theta = (theta + np.rad2deg(np.arccos((((L[0] * L[s]) + (M[0] * M[s])) + (N[0] * N[s])))))
        thetaEnd = np.abs((theta / 4.0))

# ----------------------------------------------------------------------------



        M_ENTER_P = (thetaIni / thetaSurf)
        M_EXIT_P = (thetaIni / thetaEnd)
        STOP_DIAM = self.SYSTEM.SDT[Surf].Diameter

        if (self.ApertureType == 'EPD'):
            D_Input_Pup = self.ApertureValue
        if (self.ApertureType == 'STOP'):
            D_Input_Pup = (STOP_DIAM / M_ENTER_P)

        RadPupInp = (D_Input_Pup / 2.0)
        D_Exit_Pup = (STOP_DIAM * M_EXIT_P)
        RadPupOut = (D_Exit_Pup / 2.0)

        (Xs, Ys, Zs, Ls, Ms, Ns) = RP.pick(0)
        (Xe, Ye, Ze, Le, Me, Ne) = RP.pick((- 1))

        delta_Z = 0
        ZZ = (Ls, Ms, Ns, Xs, Ys)
        v0 = scipy.optimize.fsolve(R_RMS, delta_Z, args=ZZ)

        ZZ = (Le, Me, Ne, Xe, Ye)
        vf = scipy.optimize.fsolve(R_RMS, delta_Z, args=ZZ)

        PosPupInp = np.asarray([0, 0, v0[0]])
        OPS = np.asarray([Le[0], Me[0], Ne[0]])
        DirPupSal = OPS

        (X, Y, Z, L, M, N) = RP.pick((- 1))
        px = (((L[0] / N[0]) * vf) + X[0])
        py = (((M[0] / N[0]) * vf) + Y[0])
        pz = vf

        PosPupOut = np.asarray([px[0], py[0], (vf[0] + Ze[0])])
        PosPupOutFoc = np.asarray([px[0], py[0], pz[0]])

        self.SYSTEM.Vignetting(0)
        RP.clean()
        self.RadPupInp = RadPupInp
        self.PosPupInp = PosPupInp
        self.RadPupOut = RadPupOut
        self.PosPupOut = PosPupOut
        self.PosPupOutFoc = PosPupOutFoc
        self.DirPupSal = DirPupSal
        self.menter = M_ENTER_P
        Prx = self.SYSTEM.Parax(self.W)
        self.SistemMatrix, self.S_Matrix, self.N_Matrix, self.a, self.b, self.c, self.d, self.EFFL, self.PPA, self.PPP, self.CC, self.N_Prec, self.DD = Prx
        self.FocusAiryRadius=1.22*self.W*self.EFFL/(2.0*self.RadPupInp)






    def __patern_rect(self, x, y, kx, ky):
        """__patern_rect.

        Parameters
        ----------
        x :
            x
        y :
            y
        kx :
            kx
        ky :
            ky
        """
        for i in range((- (self.Samp * kx)), ((kx * self.Samp) + 1)):
            for j in range((- (self.Samp * ky)), ((ky * self.Samp) + 1)):
                x_0 = (i / self.Samp)
                y_0 = (j / self.Samp)
                r = np.sqrt(((x_0 * x_0) + (y_0 * y_0)))
                if (r <= 1.0):
                    x.append(x_0)
                    y.append(y_0)
        return (x, y)

    def Pattern(self):
        """Pattern.
        """
        self.Samp = int(self.Samp)
        x = []
        y = []
        if (self.Ptype == 'rtheta'):
            x.append((self.rad * np.cos(np.deg2rad(self.theta))))
            y.append((self.rad * np.sin(np.deg2rad(self.theta))))
        if (self.Ptype == 'chief'):
            x.append(0.0)
            y.append(0.0)
        if (self.Ptype == 'hexapolar'):
            x.append(0.0)
            y.append(0.0)
            for j in range(1, (self.Samp + 1)):
                m = (1.0 / self.Samp)
                r = (m * j)
                k = (j * 6)
                ang = (360.0 / k)
                for h in range(0, k):
                    x.append((r * np.cos(np.deg2rad((h * ang)))))
                    y.append((r * np.sin(np.deg2rad((h * ang)))))
        if (self.Ptype == 'square'):
            (x, y) = self.__patern_rect(x, y, 1, 1)
        if (self.Ptype == 'fanx'):
            (x, y) = self.__patern_rect(x, y, 1, 0)
        if (self.Ptype == 'fany'):
            (x, y) = self.__patern_rect(x, y, 0, 1)
        if (self.Ptype == 'fan'):
            (x, y) = self.__patern_rect(x, y, 1, 0)
            (x, y) = self.__patern_rect(x, y, 0, 1)
        if (self.Ptype == 'rand'):
            p = 1000000.0
            self.Sample = ((4 * self.Samp) * self.Samp)
            x_i = (np.random.randint((- p), p, self.Sample) / p)
            y_i = (np.random.randint((- p), p, self.Sample) / p)
            for i in range(0, self.Sample):
                x_0 = x_i[i]
                y_0 = y_i[i]
                r = np.sqrt(((x_0 * x_0) + (y_0 * y_0)))
                if (r < 1.0):
                    x.append(x_0)
                    y.append(y_0)
        self.Cordx = np.asarray(x)
        self.Cordy = np.asarray(y)

    def Pattern2Field(self):
        """Pattern2Field.
        """
        self.Pattern()
        x = (self.Cordx * self.RadPupInp)
        y = (self.Cordy * self.RadPupInp)
        (Px, Py, Pz) = self.PosPupInp
        if (self.FieldType == 'angle'):
            if (self.AtmosRef == 1):
                T = self.T
                P = self.P
                H = self.H
                xc = self.xc
                lat = self.lat
                h = self.h
                l1 = self.l1
                l2 = self.l2
                z0 = self.z0
                at = Observatory()
                n1 = at.n_tph(l=l1, T=T, p=P, RH=H, xc=xc)
                n2 = at.n_tph(l=l2, T=T, p=P, RH=H, xc=xc)
                rho = at.rho(p=P, T=T, RH=H, xc=xc)
                disp = dispersion(lat, h)
                disp.setReducedHeight(P, rho)
                f_x = (z0 + self.FieldX)
                f_y = self.FieldY
                Z0 = np.sqrt(((f_x ** 2) + (f_y ** 2)))
                atm_dispersion = disp.cassini(n1, n2, Z0)
                theta = np.arctan2(f_x, f_y)
                tx = (self.FieldX + (atm_dispersion * np.sin(theta)))
                ty = (self.FieldY + (atm_dispersion * np.cos(theta)))
                shiftX = (Pz * np.tan(np.deg2rad((- tx))))
                shiftY = (Pz * np.tan(np.deg2rad((- ty))))
            else:
                shiftX = (Pz * np.tan(np.deg2rad((- self.FieldX))))
                shiftY = (Pz * np.tan(np.deg2rad((- self.FieldY))))
            f_type = 1.0
        else:
            shiftX = (- self.FieldX)

            shiftY = (- self.FieldY)

            f_type = 0.0



        x0 = (np.copy((x * f_type)) + shiftX)
        y0 = (np.copy((y * f_type)) + shiftY)
        z = (np.ones_like(x) * Pz)
        z0 = 0.0*np.copy(x)
        X2 = ((x - x0) * (x - x0))
        Y2 = ((y - y0) * (y - y0))
        Z2 = ((z - z0) * (z - z0))
        S = np.sqrt(((X2 + Y2) + Z2))
        L = ((x - x0) / S)
        M = ((y - y0) / S)
        N = ((z - z0) / S)
        return (x0, y0, z0, L, M, N)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Doublet Lens Pupil Seidel"""
from numba import jit
import timeit
import pkg_resources
""" Looking for if KrakenOS is installed, if not, it assumes that
an folder downloaded from github is run"""

required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("Not installed")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos
import numpy as np

# _________________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0.0
P_Obj.Thickness = 100
P_Obj.Glass = "AIR"
P_Obj.Diameter = 30.0
P_Obj.Name = "P Obj"

# _________________________________________#

L1a = Kos.surf()
L1a.Rc = 9.284706570002484E+001
L1a.Thickness = 6.0
L1a.Glass = "N-BK7"
L1a.Diameter = 30.0
L1a.Axicon = 0

# _________________________________________#

L1b = Kos.surf()
L1b.Rc = -3.071608670000159E+001
L1b.Thickness = 3.0
L1b.Glass = "F2"
L1b.Diameter = 30

# _________________________________________#

L1c = Kos.surf()
L1c.Rc = -7.819730726078505E+001
L1c.Thickness = 9.737604742910693E+001 - 40
L1c.Glass = "AIR"
L1c.Diameter = 30

# _________________________________________#

pupila = Kos.surf()
pupila.Rc = 0
pupila.Thickness = 40.
pupila.Glass = "AIR"
pupila.Diameter = 15.0
pupila.Name = "Ap Stop"

# _________________________________________#

P_Ima = Kos.surf()
P_Ima.Rc = 0.0
P_Ima.Thickness = 0.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 20.0

# _________________________________________#

A = [P_Obj, L1a, L1b, L1c, pupila, P_Ima]
config_1 = Kos.Setup()

# _________________________________________#




Doblete = Kos.system(A, config_1)

# _________________________________________#

W = 0.6
Surf = 4
AperVal = 3
AperType = "EPD"
fieldType = "angle"

Pup = Kos.PupilCalc(Doblete, Surf, W, AperType, AperVal)

Pup.Samp = 45
Pup.Ptype = "hexapolar"
Pup.FieldY = 3.25


#_________________________________________#

x, y, z, L, M, N = Pup.Pattern2Field()
Rayos = Kos.raykeeper(Doblete)

# _________________________________________#



start = timeit.default_timer()
print("-----------------")
cont=0.0
for i in range(0, len(x)):
    pSource_0 = [x[i], y[i], z[i]]
    dCos = [L[i], M[i], N[i]]
    Doblete.Trace(pSource_0, dCos, W)
    Rayos.push()
    cont = cont + 1.0

    # print(Doblete.NP_XYZ[0:Doblete.iter+1])
    # print(Doblete.NP_S_XYZ[0:Doblete.iter+1])




stop = timeit.default_timer()
execution_time = stop - start

print("Program Executed in "+str(execution_time/(cont*5.0))+" per ray segment")
print(cont*5, " Rays segments trace d")


# 25-oct-21 Program Executed in 0.0004910416639832562 31055.0  Segmentos
# 27-oct-21 Program Executed in 0.0002906117962646915 31055.0  Segmentos
# 29-oct-21 Program Executed in 0.00018554626739655485 31055.0  Segmentos
# 29-oct-21 Program Executed in 0.0001726453898889197 31055.0  Segmentos
# 29-oct-21 Program Executed in 0.00016649437021413582 31055.0  Segmentos
# 29-oct-21 Program Executed in 0.00015742018998550996 31055.0  Segmentos
# 30-oct-21 Program Executed in 0.0001484923241024164 31055.0  Segmentos

# 31-oct-21 implementación de FastTrace
# 31-oct-21 Program Executed in 0.00013013021978017754 455.0  Segmentos
# 31-oct-21 Program Executed in 0.00011787373611336453 31055.0  Segmentos
# 31-oct-21 Program Executed in 0.00011517481751730815 31055.0  Segmentos

# cambio de np.linalg.norm a np.sqrt(sum x**2)
# 01-nov-21 Program Executed in 0.000108276663693441 31055.0  Segmentos


# _________________________________________#
# Kos.display3d(Doblete, Rayos, 0)
